#!/usr/bin/python

import os
import sys

# args
def optParse():
	global DEBUG
	from optparse import OptionParser	
	usage = "%prog [options] "
	basename = os.path.basename(sys.argv[0])
	description = """
	declarative sensor network language magic
	"""
	parser = OptionParser(usage=usage, description=description)
	parser.add_option("-d", "--debug", action="store_true",
					  dest="debug", default=False,
					  help="Print debug messages.")
	parser.disable_interspersed_args()
	options, args =  parser.parse_args()
	DEBUG = options.debug
	if len(args) < 1 or len(args) > 2:
		print 'No. of arguments (%u) not allowed' % (len(args))
		parser.print_help()
		sys.exit(2)
	return (options, args)
	
# Main
if __name__=='__main__':
	# make compiler portable
	origpath = os.path.abspath(os.path.curdir)
	selfpath = os.path.abspath(sys.argv[0])
	selfdir = os.path.dirname(selfpath)
	os.chdir(selfdir) # switch to relative path for compatibility with existing written code
	sys.path.append('./yapps2')

	# compiler options and arguments
	(options, args) = optParse()
	infilename = os.path.normpath(os.path.join(origpath,args[0]))
	if len(args) == 1:
		outdirname = os.path.join(origpath,'gen_nesc');
	else:
		outdirname = os.path.join(origpath,args[1])
	outdirname = os.path.normpath(outdirname)
	print 'Input:  ', infilename
	print 'Output: ', outdirname

	# parse
	print '--------------------------------' 
	print 'Parsing...'
	print '--------------------------------' 
	import parser # import deferred until PYTHONPATH set
	program = parser.parse(infilename)
	print 'After parse:'
	program.dumpall()

	# high-level optimization
	print '--------------------------------' 
	print 'Optimizing...'
	print '--------------------------------' 
	import highleveloptimizer as hlopt
	program = hlopt.optimize(program)
	print 'After Optimize:'
	program.dumpall()

 	# execution planner
	print '--------------------------------' 
	print 'Planning...'
	print '--------------------------------' 
	import executionplanner as planner
	(symboltable, plan, facts) = planner.plan(program)
	print 'After plan:'
	for r in plan:
		print r.__repr__() +'\n'

	# nesc generator
	import nescgenerator
	print '--------------------------------' 
	print 'Nesc to binary...'
	print '--------------------------------' 
	nescgenerator.generate(symboltable, plan, facts,
						   ['../nesc/templates','../nesc/runtime'], #location of templates
						   outdirname) #location of output
	import motetools
	motetools.generate(symboltable, plan, facts, outdirname)
