#this file creates the intermediary language clases after going through the list of clauses generated in the
#previous levels (parser and high level transformation/optimization)

from pdb import set_trace as st
import intermediary_language as il
import grammar_classes as gc
import defines as d
import copy
import nesc_interface as nesCI


intermediaryRules = []

symbol_table = nesCI.SymbolTable()

nesC_interface_rules = []

facts = []

def do_best_joins(remaining_list):
	firstiter = True
	negfirst = False
	while(len(remaining_list) > 1):
		pred_count = 0
		for p in remaining_list:
			if isinstance(p,gc.Predicate):
				pred_count += 1
		if pred_count <= 1:
			break
		
		for i,pi in enumerate(remaining_list):
			if not isinstance(pi,gc.Predicate):
				continue
			break
		for j,pj in enumerate(remaining_list[i+1:]):
			if not isinstance(pj,gc.Predicate):
				continue
			break

		# negation tests
		# test 1: if the first predicate in the rule is megated, don't include rule in final program.  (We currently are not concerned with negations cancelling previous inferences.)
		if firstiter:
			if isinstance(pi,gc.Predicate) and not isinstance(pi,il.IOperation) and pi.negation:
				# mark for now, process later
				negfirst = True
		firstiter = False
		# test 2: if either is negated Predicate, use a negated Join
		negatedjoin = False
		for px in (pi,pj):
			if isinstance(px,gc.Predicate) and not isinstance(px,il.IOperation) and px.negation:
				negatedjoin = True
		
		count = 0
		attributes = []
		indexes1 = []
		indexes2 = []
		piVars = pi.getVariables()
		pjVars = pj.getVariables()
		for vidx, v in enumerate(piVars):
			for yidx, y in enumerate(pjVars):
				if v == y:
					attributes.append(v)
					count = count + 1
					indexes1.append(vidx)
					indexes2.append(yidx)
					break

		#create index classes
		cls_index1 = []
		cls_index2 = []
		for ind in indexes1:
			cls_index1.append(il.Index(ind))
		for ind in indexes2:
			cls_index2.append(il.Index(ind))
		
		crt_join = il.IJoin(pi,pj,cls_index1,cls_index2,negatedjoin)
		remaining_list.remove(pi)
		remaining_list.remove(pj)
		#remaining_list.append(crt_join)
		remaining_list.insert(0,crt_join)
	return negfirst


def construct_intermediary_rule(rule):
	remaining = list(rule.tail)

	
	#TODO: if left hand side is @Z

	# rearrange negation to rear and check for uses of two negations (disallowed)
	lastnegation = -1 # idx of the last negated predicate
	negfirst = False
	for i,p in enumerate(remaining):
		if isinstance(p,gc.Predicate) and p.negation:
			if lastnegation == -1:
				lastnegation = i
			else:
				raise RuntimeError, 'Two or more negated terms ("%s" and "%s") must be handled manually through explicit join for efficiency' % (remaining[lastnegation],remaining[i])
	if lastnegation != -1:
		if lastnegation == 0:
			negfirst = True # still need to eliminate negation that appears first in case of redundancy
		newremaining = []
		newremaining.extend(remaining[0:lastnegation])
		newremaining.extend(remaining[lastnegation+1:])
		newremaining.append(remaining[lastnegation])
		remaining = newremaining
	
	#aggregations first
	for p in copy.copy(remaining):
		if not isinstance(p,gc.Predicate):
			continue
		for vid,v in enumerate(p.values):
			if not isinstance(v,gc.Aggregation):
				continue
			#aggregation we replace the table with an aggregation
			agg = il.IAggregation(table,v.typeName,vid)
			remaining.remove(p)
			remaining.append(agg)

	#join now
	do_best_joins(remaining)
	
	#do select now
	for pcomp in copy.copy(remaining):
		if not isinstance(pcomp,gc.CompareCon) and not isinstance(pcomp,gc.CompareVar):
			continue

		temp_comp_list = [pcomp]

		if not pcomp in remaining:
			continue
		remaining.remove(pcomp)
		temp_var_list = copy.copy(pcomp.getVariables())

		#search for all the other comparisons that use some of thsese variables
		#we could skip this step
		# DCC this should go in highleveloptimizer e.g. canonicalizer
		for pcomp2 in copy.copy(remaining):
			if not isinstance(pcomp2,gc.CompareCon) and not isinstance(pcomp2,gc.CompareVar):
				continue
			#get a list of all related (issue a larger select)
			for vpc2 in pcomp2.getVariables():
				if vpc2 in temp_var_list:
						temp_comp_list.append(pcomp2)
						remaining.remove(pcomp2)
						for v in pcomp2.getVariables():
							if not v in temp_var_list:
								temp_var_list.append(v)
						break

					
		#search for all the predicates that have these attributes and join them
		temp_join_list = []
		for p2 in copy.copy(remaining):
			if not isinstance(p2,gc.Predicate):
				continue
			if not p2 in remaining:
				continue
			#get a list of all related (issue a larger select)
			for vp2 in p2.getVariables():
				if vp2 in temp_var_list and p2 not in temp_join_list:
						temp_join_list.append(p2)
# 						print 'moving', p2, 'from remaining to temp_join_list'
						remaining.remove(p2)
		do_best_joins(temp_join_list)

		#now do the select
		attribute_indexes = []
		comparison_operators = []
		values = []
		select_upon = temp_join_list[0]
		if len(temp_join_list) > 1:
			raise RuntimeError, 'not join right for select'
		#first identify the attributes
		for comp in temp_comp_list:
			if isinstance(comp,gc.CompareCon):
				lvar = comp.var
				rval = il.Constant(comp.con)
			else:
				lvar = comp.vars[0]
				rval = il.Index(select_upon.getVariables().index(comp.vars[1]))
			
			attribute_indexes.append(il.Index(select_upon.getVariables().index(lvar)))
			comparison_operators.append(comp.op)
			values.append(rval)
		select = il.ISelect(select_upon,attribute_indexes,comparison_operators,values)
		remaining.append(select)
	
	#take out function assignments
	
	

	#check aggregation
	last_operation = None
	for lj in remaining:
		if isinstance(lj,gc.Predicate) and not isinstance(lj,gc.Evaluation): 
			last_operation = lj
	if last_operation == None:
		raise RuntimeError, 'no outer join??'
		
	right_vars = last_operation.values
	
	remaining.remove(last_operation)
		
	#now do the selections based on functions
	for f in copy.copy(remaining):
		if isinstance(f,gc.Evaluation):
			is_in_right_hand = False
			for i,v in enumerate(right_vars):
				if str(v) == f.var:
					is_in_right_hand = True
					break
			if is_in_right_hand:
				#we have to issue a new select
				indexes = [il.Index(i)]
				param_indexes = []
				for attrn in f.func.values:
					for i,v in enumerate(right_vars):
						if str(attrn) == str(v):
							param_indexes.append(il.Index(i))
							break
				rightvals = [il.IFunction(f.func,param_indexes)]
				opers = f.op
				sel = il.ISelect(last_operation,indexes,opers,rightvals)			
				last_operation = sel
				remaining.remove(f)
		else:
			raise RuntimeError, 'remaining something else besides functions: %s' % (f)
	
	
	#do the left hand side aggregations
	for v in rule.head.values:
		if not isinstance(v,gc.Aggregation):
			continue
		#now we have an aggregation
		#the id is in the left hand side
		found = False
		for v2idx,v2 in enumerate(right_vars):
			if v2.value == v.value:
				found = True
				break
		if not found:
			raise RuntimeError, 'not found aggregation variable'
		agg = il.IAggregation(last_operation,v.typeName,v2idx)
		last_operation = agg
	
	added_columns_idx = []
	added_columns = []
	#the function project
	for f in copy.copy(remaining):
		if isinstance(f,gc.Evaluation):
			is_in_left_hand = False
			idx_in_left = -1
			for i,v in enumerate(rule.head.values):
				if str(v) == f.var:
					is_in_left_hand = True
					idx_in_left = i
					break
			if not is_in_left_hand:
				print f
				raise RuntimeError, 'unkown evaluation left'
			param_indexes = []
			for attrn in f.func.values:
				if isinstance(attrn,gc.Constant):
					param_indexes.append(il.Constant(attrn.value))
					continue
				for i,v in enumerate(right_vars):
					if str(attrn) == str(v):
						param_indexes.append(il.Index(i))
						break
			added_columns_idx.append(idx_in_left)
			added_columns.append([il.IFunction(f.func,param_indexes)])
			remaining.remove(f)
			
	if len(remaining) > 0:
		raise  RuntimeError, 'something still remaining after all ...?'
	
	#final project
	#identify the variable indexes
	right_vars = last_operation.getVariables()
	indexes = []
	for i,v in enumerate(rule.head.getVariables()):
		if not v in right_vars and i not in added_columns_idx:
			print v, " not in ",right_vars
			print 'xxx', added_columns_idx, ' ',i
			raise  RuntimeError, 'variable not in right vars for final projection'
		if v in right_vars:
			indexes.append(il.Index(right_vars.index(v)))
		else:
			indexes.append(added_columns[added_columns_idx.index(i)])
		
	outer_project = il.IProject(last_operation,indexes)
	ir = il.IRule(rule.head,outer_project);
	ir.setOriginalRuleNo(rule.original_rule_no)

	print rule, 'end'
	print outer_project
	
	return ir, negfirst

def construct_intermediary(ruleList):
	#check every rule
	for p in ruleList:
		ir, negfirst = construct_intermediary_rule(p)
		if negfirst:
			print 'skipping', p, '(negated first predicate)'
		else:
			intermediaryRules.append(ir)
		print

def classify(string_type):
	if(string_type == "uint8_t"):
		return [nesCI.TYPE_UINT8,1]
	if(string_type == "uint16_t"):
		return [nesCI.TYPE_UINT16,2]
	if(string_type == "uint32_t"):
		return [nesCI.TYPE_UINT32,4]
	if(string_type == "int8_t"):
		return [nesCI.TYPE_INT8,1]
	if(string_type == "int16_t"):
		return [nesCI.TYPE_INT16,2]
	if(string_type == "int32_t"):
		return [nesCI.TYPE_INT32,4]
	if(string_type == "char"):
		return [nesCI.TYPE_CHAR8,1]
	print 'unknown type: %s ' % (string_type)
	raise RuntimeError, 'we don\'t deal with non primitive types at the moment (we have to build equality tests into our structure)'


temp_id = 0

def newTemporary():
	global temp_id
	# dav to luci: look here to do dimensions trick with temporaries.  in particular, materialize.cardinality should be some number.
	newt = nesCI.Predicate("temp_"+str(temp_id),primary=False,materialize=None,pkey=None,host=0,replacepk=False,priority=d.DEFAULT_PRIORITY,everTriggers=True)
	newt.setTemporary()
	symbol_table.addSymbol(newt)
	temp_id = temp_id+ 1
	return newt

	
def issueRulePlan(crt_rule,iop):
	
	#first issue operations for what is below
	if isinstance(iop,il.IJoin):
		left = issueRulePlan(crt_rule,iop.first)
		right = issueRulePlan(crt_rule,iop.second)
	if isinstance(iop,il.ISelect) or isinstance(iop,il.IProject) or isinstance(iop,il.IAggregation):
		left = issueRulePlan(crt_rule,iop.table)
	
	#now issue operations for the current 
	#it is just a predicate
	if not isinstance(iop,il.IOperation):
		return symbol_table.lookupPredicateByName(iop.typeName)
	
	
	#now we have an operation
	res = None
	operation = None
	if isinstance(iop,il.IOperation):
		res = newTemporary()
	if isinstance(iop,il.IJoin):
		if right.temporary:
			operation = nesCI.EquiJoinOperation(right,left,iop.attribute_indexes2,iop.attribute_indexes1,res,iop.negation)
			res.columns.extend(right.columns)
			res.columns.extend(left.columns)
		else:
			operation = nesCI.EquiJoinOperation(left,right,iop.attribute_indexes1,iop.attribute_indexes2,res,iop.negation)
			res.columns.extend(left.columns)
			res.columns.extend(right.columns)
	if isinstance(iop,il.ISelect):
		operation = nesCI.SelectOperation(left,iop.attribute_indexes,iop.select_operators,iop.right_values,res)
		res.columns.extend(left.columns)
	if isinstance(iop,il.IProject):
		operation = nesCI.ProjectOperation(left,iop.attribute_indexes,res)
#  		res.columns.extend(left.columns) # NOT USED
	if isinstance(iop,il.IAggregation):
		operation = nesCI.AggregationOperation(left,iop.aggregation_type,iop.attribute_index,res)
# 		res.columns.extend(left.columns) FIX THIS

	res.host_idx = left.host_idx
	if operation != None:
		crt_rule.append(operation)
	else:
		raise RuntimeError, 'Operation must be valid!'

	#if its just a predicate just return it
	return res

# these better be consistent with the grammar
MAT_DEFAULT_TIMEOUT = 'infinite'
MAT_DEFAULT_CARDINALITY = 10
MAT_DEFAULT_EVICTPOLICY = 'evict_random'

def buildMaterialize(materialize):
	if materialize == None: # build default materializer
		return nesCI.NescMaterialize(MAT_DEFAULT_TIMEOUT, MAT_DEFAULT_CARDINALITY, MAT_DEFAULT_EVICTPOLICY)
	else:
		assert isinstance(materialize,gc.Materialize), st()
		return nesCI.NescMaterialize(materialize.timeout, materialize.cardinality, materialize.evictpolicy)
	
def construct_nesC_interface(program):
	#construct the initial symbol table first
	for s in program.predicateTypes:
		nescmat = buildMaterialize(s.materialize)
		priority = d.DEFAULT_PRIORITY
		if s.priority != None:
			priority = s.priority
		symbol = nesCI.Predicate(s.name,s.builtin,nescmat,s.pkey,s.host_idx,s.replacepk, priority, s.everTriggers)
		for i,v in enumerate(s.values):
			l = classify(v)
			col = nesCI.Column(l[0],l[1])
			#print s.name+' colid '+str(i)+' type '+str(l[0])
			col.strtypename = v
			symbol.addColumn(col)
		symbol_table.addSymbol(symbol)
		#facts support
		for f in program.facts:
			if f.p.type == s:
				#create the two lists
				compares = []
				consts = []
				for comp in f.compares:
					if isinstance(comp,gc.CompareCon):
						if comp.var != str(f.p.values[0]):
							print 'var: '+str(comp.var)+' host: '+str(f.p.values[0])
							raise RuntimeError, 'facts suport only comparisons on the host field'
						compares.append(comp.op)
						consts.append(comp.con)
					else:
						raise RuntimeError, 'bad comparison for facts (not CompareCon)',comp
				newfact = nesCI.TableFact(symbol,f.p.values,compares,consts)
				facts.append(newfact)
	
	#add functions
	for f in program.functionTypes:
		l = classify(f.retType)
		ret = nesCI.Column(l[0],l[1])
		function = nesCI.Function(f.name,ret,f.builtin)
		for v in f.values:
			l = classify(v)
			function.addParameter(nesCI.Column(l[0],l[1]))
		symbol_table.addSymbol(function)
	
	#now construct the rules
	#for r in intermediaryRules:
	for r in intermediaryRules:
		#print str(r.left.values[r.left.host]) +'<-'+ str(r.right.values[r.right.host])
		network_broadcast = False
		if not r.left.values[r.left.host] == r.right.values[r.right.host]:
			network_broadcast = True
		rule_nesC = nesCI.Rule([], network_broadcast)

		res = r.left
		lastres = issueRulePlan(rule_nesC.operations,r.right)
		
		#add the final one (remove the temporary symbol and replace it with r.left)
		newres = symbol_table.lookupPredicateByName(res.typeName)
		symbol_table.removePredicateByName(lastres.name)
		rule_nesC.operations[len(rule_nesC.operations)-1].result = newres
		rule_nesC.setOriginalRuleNo(r.original_rule_no)
		
		nesC_interface_rules.append(rule_nesC)
		

def duplicate_rules(ruleList):
	#check every rule
	no_orig_rules = len(ruleList)
	crt = 0
	while crt < no_orig_rules:
		r = ruleList[crt]
		r.setOriginalRuleNo(crt)
		n = len(r.tail)
		replace_original = r.nontriggers[0]
		
		if not replace_original:
			r.tail[0].type.everTriggers = True
		i = 1
		#print 'for rule ',r,' non triggers are ',r.nontriggers
		while i < n:
			if not isinstance(r.tail[i],gc.Predicate) or r.nontriggers[i]:
				i = i + 1
				continue;
			neworder = copy.copy(r.tail)
			tmp = neworder[0]
			neworder[0] = neworder[i]
			neworder[i] = tmp
			newrule = gc.Rule(r.head,neworder)
			newrule.setOriginalRuleNo(crt)
			if replace_original:
				replace_original = False
				ruleList[crt] = newrule
			else:
				ruleList.append(newrule)
			newrule.tail[0].type.everTriggers = True
			i = i+1
		crt = crt + 1		


def plan(program):

	#what this does is duplicate the rules such that each predicate is first 
	#actually we realized we need multiple execution plans after the construction of the
	#tuple by tuple backend and this was the easiest way out :)
	duplicate_rules(program.rules)

	construct_intermediary(program.rules)
	
	construct_nesC_interface(program)
	 
	return (symbol_table, nesC_interface_rules, facts)
