	
import defines as defs
import grammar_classes as gc

%%
parser Sdlog:
        ignore:         r'\s+'
        token END:      "$"
        token PER:      "\."
        token QUE:      "\?"
        token IMP:      "\:\-"
        token HSHIMP:   "\#\:\-"
        token IMPHSH:   "\:\-\#"
        token ATT:      "\@"
        token ALL:      "\*"
        token NUM:      r'[0-9]+'
        token CON:      r'[0-9]+'
		#r'[a-z][a-zA-Z0-9_]*'
        token VAR:      r'[A-Z][a-zA-Z0-9_]*'
        token DC:       '\_'
        token PRE:      r'[a-zA-Z_][a-zA-Z0-9_]*'
        token FUN:      r'f_[a-zA-Z0-9_]*'
        token AGG:      r'AVG|MIN|MAX|MEDIAN'
        token HLP:      '\!'
        token COM:      r'\%[-+*/~#!@%^|&?="\'.:,()<>\sa-zA-Z0-9_\\]*'
        token HSH:      "\#"
        token EVL:      r"="
        token CMP:      r"==|\!=|\>[=]*|\<[=]*"
        token TYPEDEF:  "typedef"
        token BUILTIN:  "builtin"
        token MATERIAL: "materialize"
        token IMPORT:   "import"
        token CTYPE:    r'[a-zA-Z_][a-zA-Z0-9_]*'
        token PKY:      r'\^'
        token STR:      r'[a-zA-Z_][-a-zA-Z0-9_./]*'
        token NESCCOMPONENT:    r'[a-zA-Z_][a-zA-Z0-9_]*'
        token MAT_TIMEOUT:      r'[1-9][0-9]*|infinite'
        token MAT_CARDINALITY:  r'[1-9][0-9]*'
        token MAT_EVICTPOLICY:  r'evict_random|evict_deny|evict_oldest|evict_newest'
        token PRIORITY: "priority"
        token NONTRIGGER: r'~'
        token NOT:      r'\-'

# dcc Todo and Known Issues
# - The grammar is currently limited in its selection expressivity.
# Only ==, <, > are currently supported.
# - Currently no support for arithmetic operations in evaluations.
# E.g. the term X=Y+Z is not allowed.

        rule clause:
        predicate 
        clausetail<<predicate>> 
        END                 {{ return clausetail }}
    |   HLP help END        {{ return help }}
    |   COM END             {{ return None }}
    |   END                 {{ return None }}

        rule clausetail<<head>>:
        PER                 {{ return gc.Fact(head) }}
    |   "\," factcompares PER {{ return gc.Fact(head,factcompares) }}
    |   QUE                 {{ return gc.Query(head) }}
    |   IMP body PER  {{ x = gc.Rule(head, body); return x }}

# are these enough to provide push/pull hints?
#    |   IMP body PER       {{ return ["agnostic", body] }}
#    |   HSHIMP body PER    {{ return ["pull", body] }}
#        IMPHSH body PER    {{ return ["push", body] }}

        rule factcompares:
        compare            {{ e = [compare] }}
        ("\," compare       {{ e.append(compare) }} )* {{ return e }}
            
        rule compare:
        VAR CMP comparerhs<<CMP,VAR>> {{ return comparerhs }}

        rule body: 
        term                {{ e = [term] }}
        ( "\,"  term        {{ e.append(term); }})* {{ return e }}
        
        rule term:
        predicate               {{ return predicate }}
#    |   NOT predicate           {{ return gc.Not(predicate) }}
    |   VAR nonpredicate<<VAR>> {{ return nonpredicate }}

        rule predicate:
                            {{ negation = False }}
        ( NOT               {{ negation = True }} )*
        PRE tuple           {{ trigger = True }}
        ( NONTRIGGER        {{ trigger = False }} )* # spec too loose e.g. ~~
                            {{ return gc.Predicate(PRE, tuple, trigger, negation) }}
 
        rule nonpredicate<<var>>:
        CMP comparerhs<<CMP,var>> {{ return comparerhs }}
    |   EVL function        {{ return gc.Evaluation(EVL, var, function) }}
        
        rule comparerhs<<cmp,lhs>>:
        VAR                 {{ return gc.CompareVar(cmp,lhs,VAR) }}
    |   CON                 {{ return gc.CompareCon(cmp,lhs,CON) }}

        rule tuple:
        r"\("               {{ l = [] }}
        ( arg               {{ l.append(arg) }}
            ( "\," arg      {{ l.append(arg) }} )* ) 
        r"\)"               {{ return l }}

        rule arg:
        CON                 {{ return gc.Constant(CON) }}
    |   DC                  {{ return gc.DontCare() }}
    |   VAR                 {{ return gc.Variable(VAR) }}
#    |   ATT VAR             {{ return gc.HostAt(gc.Variable(VAR)) }} # previously only allowed at sign for variables, not constants
    |   ATT athost          {{ return athost }}
    |   "\<" AGG 
        "\," VAR "\>"       {{ return gc.Aggregation(AGG, VAR) }}

        rule athost:
        CON                 {{ return gc.HostAt(gc.Constant(CON)) }}
    |   VAR                 {{ return gc.HostAt(gc.Variable(VAR)) }}
    |   ALL                 {{ return gc.HostAt(gc.Constant(defs.BCAST_ADDR)) }}
# todo: 65535 here is dirty, ties into nescgenerator.  but it was quicker

        rule function:
        FUN tuple           {{ return gc.Function(FUN, tuple) }} 
        # above rule is shortcut for now.
        # should restrict args of function to only VARs

        rule help:
        hlpTypedef PER      {{ return hlpTypedef }}
    |   hlpBuiltin PER      {{ return hlpBuiltin }}
    |   hlpMaterialize PER  {{ return hlpMaterialize }}
    |   hlpImport PER       {{ return hlpImport }}
    |   hlpPriority PER     {{ return hlpPriority }}
        
      
        rule hlpTypedef:
        TYPEDEF r"\("       {{ l = []; replacepk = False }}
        ( (HSH              {{ replacepk = True }})*
            PRE
            ( "\," tdparam  {{ l.append(tdparam) }} )* ) 
        r"\)"               {{ return gc.PredicateType(PRE, l, replacepk=replacepk) }}

        rule tdparam:
        PKY CTYPE           {{ return gc.PrimaryKeyMember(CTYPE) }}
    |   CTYPE               {{ return CTYPE }}
        
        rule hlpBuiltin:
        BUILTIN r"\("
        PRE "\," NESCCOMPONENT
        r"\)"               {{ return gc.Builtin(PRE, NESCCOMPONENT) }}
        
        rule hlpMaterialize:
        MATERIAL r"\("
        PRE "\," MAT_TIMEOUT "\," MAT_CARDINALITY "\," MAT_EVICTPOLICY
        r"\)"               {{ return gc.Materialize(PRE, MAT_TIMEOUT, MAT_CARDINALITY, MAT_EVICTPOLICY) }}

        rule hlpPriority:        
        PRIORITY r"\("
        PRE "\," NUM r"\)"  {{ return gc.Priority(PRE, NUM) }}

        rule hlpImport:
        IMPORT "\("         {{ l = [] }}
        ( qstring           {{ l.append(qstring) }}
            ("\," qstring   {{ l.append(qstring) }} )* )
        "\)"                {{ return gc.Import(l)  }}

        rule qstring:
        "\'" STR "\'"       {{ return STR }}
    |   "\"" STR "\""       {{ return STR }}
