import sys
from pdb import set_trace as st	
# Helper funcs

def seqstr(seq):
	ret = ''
	sep = ','
	for s in seq:
		ret = ret + str(s) + sep
	return ret[0:(len(ret)-len(sep))]

def resolveType(type1,type2):
	if type1 == type2:
		return (False, type1)
	elif type1 == None:
		return (True, type2)
	elif type2 == None:
		return (True, type1)
	else:
		st()
		raise RuntimeError, 'Type inconsistency: %s != %s' % (type1, type2)

# Classes

class Program:
	def __init__(self, rules, facts, queries, predicateTypes, functionTypes, builtins, materializes):
		self.rules = rules
		self.facts = facts
		self.queries = queries
		self.predicateTypes = predicateTypes
		self.functionTypes = functionTypes
		self.builtins = builtins
		self.materializes = materializes
# 	def __repr__(self):
	def dumpall(self):
		print 'Rules'
		self._dumppart(self.rules)
		print 'Facts'
		self._dumppart(self.facts)
		print 'Queries'
		self._dumppart(self.queries)
		print 'Predicate Types'
		self._dumppart(self.predicateTypes)
		print 'Function Types'
		self._dumppart(self.functionTypes)
	def _dumppart(self,part):
		if len(part) == 0: print '[none]'
		else:
			for p in part: print p
		print

class Import:
	def __init__(self,imports):
		self.imports = imports
	
	

class DontCare:
	def __init__(self):
		pass
	def __repr__(self):
		return "_"
		
class Constant:
	def __init__(self, value):
		self.value = value
	def __repr__(self):
		return "%s" % (self.value)

class HostAt:
	def __init__(self, var):
		self.var = var
	def __repr__(self):
		return "@%s" % (self.var)

class PrimaryKeyMember:
	def __init__(self, ctype):
		self.ctype = ctype
	def __repr__(self):
		return "#%s" % (self.ctype)

class Variable:
	def __init__(self, value):
		self.value = value
	def __repr__(self):
		return "%s" % (self.value)

class VariableContainer: # fake an Java-like interface
	def getVariables(self): raise NotImplementedError
	def getVariableType(self,vidx): raise NotImplementedError
	def setVariableType(self,vidx,newType): raise NotImplementedError
	
class Aggregation:
	def __init__(self, typeName, value):
		self.typeName = typeName
		self.value = value
#		 self.type = AggregactionType.lookup(self)
	def __repr__(self):
		return "<%s, %s>" % (self.typeName, self.value)
#	 def getVariables(self):
#		 return [self.value.value]
#	 def getVariableType(self, vidx):
#		 return self.type.value[vidx]
#	 def setVariableType(self, vidx, newType):
#		 self.type.value[vidx] = newType

# class AggregationType:
#	 def __init__(self, name, varType):
#		 self.name = name
#		 self.varType = varType

class CompareCon(VariableContainer):
	def __init__(self, op, var, con):
		self.op = op
		self.var = var
		self.con = con
		self.type = [None]
# 	def get_values(self):
# 		return [self.var]
# 	def set_values(self, value):
# 		self.var = value
# 	values = property(get_values, set_values)
	def __repr__(self):
		return "%s%s%s" % (self.var, self.op, self.con)
	def getVariables(self):
		return [self.var]
	def getVariableType(self,vidx):
		return self.type[vidx]
	def setVariableType(self,vidx,newType):
		self.type[vidx] = newType
	def isAssigned(self):
		if self.type[0] == None:
			print '%s not assigned in %s' % (self.var, self)
			return False
		else:
			return True

class CompareVar(VariableContainer):
	def __init__(self, op, var1, var2):
		self.op = op
		self.vars = [var1,var2]
		self.types = [None, None]
# 	def get_values(self):
# 		return self.vars
# 	def set_values(self, value):
# 		self.var = value		
# 	values = property(get_values, set_values)
	def __repr__(self):
		return "%s%s%s" % (self.vars[0], self.op, self.vars[1])
	def getVariables(self):
		return self.vars
	def getVariableType(self,vidx):
		return self.types[vidx]
	def setVariableType(self,vidx,newType):
		self.types[vidx] = newType
		otheridx = (vidx + 1) % 2
		if self.types[otheridx] == None:
			self.types[otheridx] = newType
		elif self.types[otheridx] != newType:
			raise RuntimeError, 'Compare of inconsistent types'
	def isAssigned(self):
		flag = True
		for t in self.types:
			if t == None:
				print '%s not assigned in %s' % (self.vars[self.types.index(t)],self)
				flag = False
		return flag

class Evaluation(VariableContainer):
	def __init__(self, op, var, func):
		self.op = op
		self.var = var
		self.func = func
	def __repr__(self):
		return "%s%s%s" % (self.var, self.op, self.func)
	def getVariables(self):
		return [self.var] + self.func.getVariables()   
	def getVariableType(self,vidx):
		if vidx == 0:
			return self.func.getReturnType()
		else:
			return self.func.getVariableType(vidx-1)
	def setVariableType(self,vidx,newType):
		if vidx == 0:
			self.func.setReturnType(newType)
		else:
			self.func.setVariableType(vidx-1,newType)
	def isAssigned(self):
		return self.func.isAssigned()

class Function:
	def __init__(self, typeName, values):
		self.typeName = typeName
		self.values = values
		self.type = FunctionType.lookup(self)
	def __repr__(self):
		return "%s(%s)" % (self.typeName, seqstr(self.values))
	def getVariables(self):
		l = []
		for v in self.values:
			if isinstance(v,Variable):
				l.append(v.value)
		return l
	def getReturnType(self):
		return self.type.retType
	def setReturnType(self, newType):
		self.type.retType = newType
	def getVariableType(self,vidx):
		aidx = self.findVariable(vidx) #idx into all, not just variables
		return self.type.values[aidx]
	def setVariableType(self,vidx,newType): 
		aidx = self.findVariable(vidx)
		self.type.values[aidx] = newType
# 		if FunctionType.typess[self.type.name] != self.type:
# 			st()
# 			print FunctionType.typess[self.type.name],":", self.type:
# 			FunctionType.typess[self.type.name] = self.type
	def findVariable(self,vidx):
		j = 0
		for i,v in enumerate(self.values):
			if isinstance(v,Variable):
				if vidx == j:
					return i
				j += 1
	def isAssigned(self):
		return self.type.isAssigned()
	
class FunctionType:
	typess = {}
	def __init__(self, name, var2):
		self.name = name
		self.builtin = None
		self.retType = None
		if isinstance(var2,list):
			self.values = var2
			raise RuntimeError, "Unused, unimplemented typedef"
		elif isinstance(var2,int):
			self.values = [None for i in range(var2)]
		else:
			raise RuntimeError, "Illegal assignment to var2"
		FunctionType.typess[name] = self
		self.allTypes = FunctionType.typess
	def isAssigned(self):
		if self.builtin == None:
			print '%s has no builtin defined' % (self.name)
			return False
		for i,v in enumerate(self.values):
			if v == None:
				print 'Col %u of %s is not assigned' % (i,self.name)
				return False
		if (self.retType == None):
			print 'Rettype of %s is not assigned' % (self.name)
			return False
		return True
	def __repr__(self):
		return "%s(%s)->%s [%s]" % (self.name, seqstr(self.values), self.retType, self.builtin)
	@staticmethod
	def lookup(function):
		if function.typeName not in FunctionType.typess:
			FunctionType.typess[function.typeName] = FunctionType(function.typeName, len(function.values))
		return FunctionType.typess[function.typeName]
	@staticmethod
	def lookupByName(funcName):
		if funcName in FunctionType.typess:
			return FunctionType.typess[funcName]
		return None		

class Predicate(VariableContainer):
	def __init__(self, typeName, values, istrigger, negation):
		self.typeName = typeName
		self.values = values
		self.host = self.getHostIdx(values)
		self.values[self.host] = self.values[self.host].var
		self.type = PredicateType.lookup(self)
		assert len(values) == len(self.type.values), 'mismatch of variables in %s: instance has %s (len=%u) whereas type has %s (len=%u)' % (typeName, values, len(values), self.type.values, len(self.type.values))
		self.type.host_idx = self.host
		# special check to support efficient timer allocations
		if self.typeName == 'timer' and isinstance(self.values[1],Constant):
			pass
# 			print 'xxxxxxx TODO efficient timer allocation'
		self.istrigger = istrigger # indicates whether this predicate ever needs to fire this rule
		self.negation = negation
	def __repr__(self):
		ret = "%s@%s(%s)" % (self.type.name, self.values[self.host], seqstr(self.values))
		if self.negation:
			ret = '%s%s' % ('-', ret)
		if not self.istrigger:
			ret = "%s%s" % (ret, '~')
		return ret
	def getHostIdx(self, values):
		cnt = 0
		for i,v in enumerate(values):
			if isinstance(v,HostAt):
				cnt += 1
				idx = i
		if cnt != 1:
			raise RuntimeError, "Too few or too many HostAt in Predicate"
		return idx
	def getVariables(self):
		l = []
		for v in self.values:
			if isinstance(v,Variable) or isinstance(v,Aggregation):
			  l.append(v.value)
		return l
	def setVariable(self, vidx, value):
		aidx = self.findVariable(vidx)
		self.values[aidx].value = value
	def getVariableType(self,vidx):
		aidx = self.findVariable(vidx) #idx into all, not just variables
		return self.type.values[aidx]
	def setVariableType(self,vidx,newType):
		aidx = self.findVariable(vidx)
		self.type.values[aidx] = newType
		if PredicateType.typess[self.type.name] != self.type:
# 			st() # TODO this is the WRONG place to add stuff but oh well - should really figure out WHERE the predicateType is spliting off from the self.type and creating a dup of the obj.
# 			print PredicateType.typess[self.type.name], " xxx ",self.type
			for i in range(len(self.type.values)):
				(currChange, tk) = resolveType(PredicateType.typess[self.type.name].values[i],self.type.values[i])
				PredicateType.typess[self.type.name].values[i] = tk
				self.type.values[i] = tk
# 			print PredicateType.typess[self.type.name], " yyy ",self.type
			PredicateType.typess[self.type.name] = self.type # reset to point to each other			
	def findVariable(self,vidx):
		j = 0
		for i,v in enumerate(self.values):
			if isinstance(v,Variable) or isinstance(v,Aggregation):
				if vidx == j:
					return i
				j += 1
	def isAssigned(self):
		return self.type.isAssigned()

class Builtin:
	def __init__(self, name, nescC):
		self.name = name
		self.nescC = nescC
	def __repr__(self):
		return "builtin(%s,%s)" % (self.name, self.nescC)

class Materialize:
	def __init__(self, name, timeout, cardinality, evictpolicy):
		self.name = name
		self.timeout = timeout
		self.cardinality = cardinality
		self.evictpolicy = evictpolicy
	def __repr__(self):
		return "materialize(%s,%s,%s,%s)" % (self.name, self.timeout, self.cardinality, self.evictpolicy)		

class PredicateType:
	typess = {}
	def __init__(self, name, var2, replacepk=False):
		self.name = name
		self.builtin = None
		self.materialize = None
		self.priority = None
		self.replacepk = replacepk
		if isinstance(var2,list):
			self.values = []
			pkey = []
			for v in var2:
				if isinstance(v,PrimaryKeyMember):
					ctype = v.ctype
					pkeymem = True
				elif isinstance(v,str):
					ctype = v
					pkeymem = False
				else:
					raise Exception, 'unsupported instance(%s) type(%s) in PredicateType initialization' % (v, type(v))
				self.values.append(ctype)
				pkey.append(pkeymem)
			if sum(pkey) == 0: # if all false, default to pkey is entire tuple
				self.pkey = [True for i in range(len(pkey))]
			else:
				self.pkey = pkey
		elif isinstance(var2,int):
			self.values = [None for i in range(var2)]
			self.pkey = [True for i in range(var2)]
		else:				 
			raise RuntimeError, "Illegal assignment to var2"
# 		print name, self.replacepk
# 		st()
		assert name not in PredicateType.typess, '%s already exists in %s' % (name, PredicateType.typess)
		PredicateType.typess[name] = self
		self.allTypes = PredicateType.typess
		#the host is actually set in each predicate :))
		self.host_idx = -1
		self.everTriggers = False #we'll set it up later
	def isAssigned(self):
		flag = True
		for i,v in enumerate(self.values):
			if v == None:
				print 'Col %u of %s is not assigned' % (i,self.name)
				flag = False
		return flag
	def __repr__(self):		
		ret = "%s(%s)" % (self.name, seqstr(self.__pkeyRepr()))
		if self.materialize != None:
			ret = "%s [%s]" % (ret, self.materialize)
		if self.builtin != None:
			ret = "%s [%s]" % (ret, self.builtin)
# 		st()
		if self.replacepk == True:
			ret = "#%s" % (ret)
		if self.priority != None:
			ret = "(%i) %s" % (self.priority, ret)
		# it is invalid to have both builtin and materialize! this check is not done
		return ret
	def __pkeyRepr(self):
		ret = []
		pkeyinv = [not pkeymem for pkeymem in self.pkey]
		if sum(pkeyinv) == 0:
			return self.values
		for i,val in enumerate(self.values):
			if self.pkey[i] == True: ret.append('^'+val)
			else: ret.append(val)
		return ret
	@staticmethod
	def lookup(predicate):
		if predicate.typeName not in PredicateType.typess:
			PredicateType.typess[predicate.typeName] = PredicateType(predicate.typeName, len(predicate.values))
		return PredicateType.typess[predicate.typeName]
	@staticmethod
	def lookupByName(predName):
		if predName in PredicateType.typess:
			return PredicateType.typess[predName]
		return None		

class Query:
	def __init__(self, p):
		self.p = p # predicate
	def __repr__(self):
		return "%s?" % (self.p)
	def addPredicateType(self):
		PredicateType.lookup(self.p)
	def isAssigned(self):
		return self.p.isAssigned()

class Fact:
	def __init__(self, p, compares=[]):
		self.p = p # predicate
		assert isinstance(compares,list)
		self.compares = compares
		# todo: checks on correctness of compares var with p host
	def __repr__(self):
		if len(self.compares) == 0:
			return "%s." % (self.p)
		else:
			return "%s, %s." % (self.p, seqstr(self.compares))
	def addPredicateType(self):
		PredicateType.lookup(self.p)
	def isAssigned(self):
		return self.p.isAssigned()

class Rule:
	def __init__(self, head, tail):
		self.head = head
		self.tail = tail
		self.nontriggers = [False for i in range(len(tail))]
		for i,t in enumerate(tail):
			if isinstance(t,Predicate) and not t.istrigger:
				self.nontriggers[i] = True
	def __repr__(self):
		return "%s :- %s." % (self.head, seqstr(self.tail))
	def addPredicateTypes(self):
		for p in [self.head] + self.tail:
			if isinstance(p,Predicate):
				PredicateType.lookup(p)
	def joinTypeInfer(self):
		change = False
		self.matchtbl = []
		all = [self.head] + self.tail
		for i,pi in enumerate(all):
			for j,pj in enumerate(all[i+1:]):
				if not isinstance(pi,VariableContainer) or not isinstance(pj,VariableContainer):
					continue
				piVars = pi.getVariables()
				pjVars = pj.getVariables()
				for vidx, v in enumerate(piVars):
					for yidx, y in enumerate(pjVars):
						if v == y:
							ti = pi.getVariableType(vidx)
							tj = pj.getVariableType(yidx)
							(currChange, tk) = resolveType(ti,tj)
#							 print currChange, v, y, ti, tj, tk, pi, pj
							pi.setVariableType(vidx,tk)
							pj.setVariableType(yidx,tk)
							matchentry = MatchEntry(v, pi, vidx, pj, yidx)
							self.matchtbl.append(matchentry)
							if currChange:
								change = True								
		return change
	def tailmatchtbl(self):
		excludelist = [self.head]
		return self.excludematchtbl(excludelist)
	def excludematchtbl(self, excludelist):
		tmpmatchtbl = []
		for m in self.matchtbl:
			if m.predL in excludelist or m.predR in excludelist:
				continue
			tmpmatchtbl.append(m)
		return tmpmatchtbl
	def isAssigned(self):
		all = [self.head] + self.tail
		for term in all:
			if isinstance(term,VariableContainer) and not term.isAssigned():
				return False
		return True
	def setOriginalRuleNo(self,original_rule_no):#to support multiple execution plans
		self.original_rule_no = original_rule_no

class MatchEntry:
	def __init__(self,var,predL,predLidx,predR,predRidx):
		self.var = var
		self.predL = predL
		self.predLidx = predLidx
		self.predR = predR
		self.predRidx = predRidx
	def __repr__(self):
		return "%s" % ([self.var,self.predL,self.predLidx,self.predR,self.predRidx])	

class Priority:
	def __init__(self, name, priority):
		self.name = name
		self.priority = int(priority)
	def __repr__(self):
		return "priority(%s,%i)" % (self.name, self.priority)
