#!/usr/bin/python

import copy
import grammar_classes as gc
from pdb import set_trace as st

# Equality Selection Rewrite
def replaceEqualitySelectionByEquijoin(r,compareterm):
	comparevars = compareterm.getVariables()

	all = [r.head] + r.tail
	
	for t in all:
		if t == compareterm:
			continue
		vs = t.getVariables()
		for i,v in enumerate(vs):
			if v == comparevars[1]:
				t.setVariable(i, comparevars[0])
	r.tail.remove(compareterm)
	return r

def eliminateEqualitySelection(program):
	# pass by ref changes original program
	newRules = []
	for r in program.rules:
		newRule = r
		for t in r.tail:
			if isinstance(t,gc.CompareVar) and len(t.getVariables())==2 and t.op == '==':
				print 'replacing equality selection with equijoin in: '
				print '\told:%s' % (r)
				newRule = replaceEqualitySelectionByEquijoin(r,t)
				print '\tnew:%s' % (newRule)
				

		newRules.append(newRule)
	program.rules = newRules
	return program			

# Link Local Rewrite
def linklocalRewrite(program):
	# dcc todo: rewrite link local rules for:
	# 1. maximizing sharing
	# 2. minimize (naive) message costs (actually everything here is minimizing messages
	# 3. maximize rule selectivity i.e. don't ship head if head will not participate in a rule that is activated

	# Notes on currently implemented: Rewrite order is based on order
	# of appearance (orderofoappearance).  Namely, rendevouze happens
	# at the first host to appear in the tail ('this'); the second
	# host to appear ('that') sends a msg to the first host with all
	# the predicates it can join.  Two caveats: (1) all 'link'
	# predicates are discarded and recreated and (2) all
	# non-predicates e.g. selections and special projections are run
	# at 'this'.

	newrules = []

	for r in program.rules:
		hosthash = {}
		hosthashlink = {}
		nonpreds = []
		orderofappearance = []
		potentialdiscards = []
		for t in r.tail:
			if not isinstance(t,gc.Predicate):
				nonpreds.append(t)
				continue
			# old behavior upon encountering "link": discard links
# 			if t.typeName == 'link':
# 				potentialdiscards.append('WARNING: discarding %s' % (t))
# 				continue
				# new behavior: retain links
			hostvar = t.values[t.host].value
			if hostvar in hosthash:
				hosthash[hostvar].append(t)
			else:
				hosthash[hostvar] = [t]
				orderofappearance.append(hostvar)
			# new behavior upon encountering "link": discard links
			if t.typeName == 'link':
				hosthashlink[hostvar] = True				
		if len(hosthash) == 1:
			newrules.append(r)
		elif len(hosthash) == 2:
			print 'Rewriting %s' % (r)
			for discardwarning in potentialdiscards:
				print discardwarning
			thishost = orderofappearance[0]
			thathost = orderofappearance[1]

			# create msg predicate
			thathosttypename = '_msg'
			thathostvalues = [gc.HostAt(gc.Variable(thishost))]
			predIsTrigger = False
			for p in hosthash[thathost]:
				thathostvalues.extend(p.values)
				thathosttypename = thathosttypename + '_' + p.typeName
				if p.istrigger:
					predIsTrigger = True
			predMsg = gc.Predicate(thathosttypename, thathostvalues, predIsTrigger)
			cp_predMsg = copy.deepcopy(predMsg)

			# make that rule
			if thathost not in hosthashlink:
				thatlink = gc.Predicate('link', [gc.HostAt(gc.Variable(thathost)),
																				 gc.Variable(thishost),
																				 gc.DontCare()], True)
				hosthash[thathost].append(thatlink)
			thatrule = gc.Rule(cp_predMsg, hosthash[thathost])
# 			st()
			thatrule.joinTypeInfer() #ugly
			newrules.append(thatrule)

			# make this rule
			hosthash[thishost].append(predMsg)
			hosthash[thishost].extend(nonpreds)
			if thishost not in hosthashlink:
				thislink = gc.Predicate('link', [gc.HostAt(gc.Variable(thishost)),
																				 gc.Variable(thathost),
																				 gc.DontCare()])
				hosthash[thishost].append(thislink)
			thisrule = gc.Rule(r.head, hosthash[thishost])
			thisrule.joinTypeInfer() #ugly
			newrules.append(thisrule)
			print '... to   %s' % (thatrule)
			print '         %s' % (thisrule)
			program.predicateTypes = gc.PredicateType.typess.values() #ugly
# 			st()
		else:
			raise Exception, 'Unsupported: No. of hosts is %u in rule %s' % (len(hosthash), r)

	program.rules = newrules
	return program

# Canonicalize Constants
def canonicalizeConstants(program):
	for r in program.rules:
		# process head
		if isinstance(r.head,gc.Predicate):
			for i,val in enumerate(r.head.values):
				if isinstance(val,gc.Constant):
					newVar = dontcare()
					r.head.values[i] = newVar
# 					funcname = 'f_const'
					funcname = 'f_const_%s' % (r.head.type.values[i])
					newFunc = gc.Function(funcname, [val])
					newEval = gc.Evaluation('=', newVar.value, newFunc)
					r.tail.append(newEval)

					# restore invariants, TODO and ugly!!!
					if newFunc.type.values[0] != None and newFunc.type.values[0] != r.head.getVariableType(i):
# 						st()
						print 'rule="%s", funcname="%s", type1="%s", type2="%s"' % (r,newFunc, newFunc.type.values[0], r.head.getVariableType(i))
						raise Exception, 'TODO already assigned one value to this variable function type... now trying to assign another.  should be allowed but since we do not have VariableContainers do not permit variable types, then this cannot be done for now'
					newFunc.type.values[0] = r.head.getVariableType(i)
					newFunc.type.builtin = 'FuncConstC'
# 					gc.Builtin(funcname,'dummy') # TODO not sure how to insert this.
					program.functionTypes = gc.FunctionType.typess.values() # ugly
					r.joinTypeInfer() #ugly

					
		# process tail
		for t in r.tail:
			if isinstance(t,gc.Predicate):
				for i,val in enumerate(t.values):
					if isinstance(val,gc.DontCare):
						newVar = dontcare()
						t.values[i] = newVar
					if isinstance(val,gc.Constant):
						# found const in body predicate
						newVar = dontcare()
						t.values[i] = newVar
# 						newCompare = gc.CompareCon('==', newVar, val) # TODO should be this way but old frontend assumes strings when object is the correct way to do it - too lazy to change.
						newCompare = gc.CompareCon('==', newVar.value, val.value)
						r.tail.append(newCompare)
						r.joinTypeInfer() #ugly


	return program
	

# Word align, but must deal with primitive types!, timerSetting and then type infer propagate
def wordalign(program):
	print 'Word-align all predicates and functions'
	for t in gc.PredicateType.typess.values():
		converted = False
		for i,v in enumerate(copy.copy(t.values)):
			if v == 'int8_t':
				t.values[i] = 'int16_t'
				converted = True
			elif v == 'uint8_t':
				t.values[i] = 'uint16_t'
				converted = True
		if converted:
			print '\tAligned %s' % (t)
# 	print gc.PredicateType.typess.values()
	for t in gc.FunctionType.typess.values():
		converted = False
		for i,v in enumerate(copy.copy(t.values)):
			if v == 'int8_t':
				t.values[i] = 'int16_t'
				converted = True
			elif v == 'uint8_t':
				t.values[i] = 'uint16_t'
				converted = True
		if t.retType == 'int8_t':
			t.retType = 'int16_t'
			converted = True
		elif t.retType == 'uint8_t':
			t.retType = 'uint16_t'
			converted = True
		if converted:
			print '\tAligned %s' % (t)
# 	print gc.FunctionType.typess.values()
# 	st()
	program.functionTypes = gc.FunctionType.typess.values()
	program.predicateTypes = gc.PredicateType.typess.values()
	return program


__dontcare = 0
def dontcare():
	global __dontcare
	__dontcare += 1
	return  gc.Variable('_DC' + str(__dontcare) + '_')

def optimize(program):
	p = program
# 	p = eliminateEqualitySelection(p)
	p = canonicalizeConstants(p) # run1: eliminate consts in tail
	p = linklocalRewrite(p)
	p = canonicalizeConstants(p) # run2: eliminate DCs introduced by link(X,Y,_)
	p = wordalign(p)
	
	# TODO: we need to call several sync funcs (e.g. joinTypeInfer) several times because of new stuff introduced.  it would be nice to do this once after everything is done.  grep 'ugly'
	return p

