import os
import sys
import grammar_classes as gc
import defines as d

comparison_oper_string = ["<", "<=",">", ">=", "==", "!="] # this should EXACTLY match what's in the grammar.g otherwise we're in trouble...

class IOperation(gc.Predicate):
	def __init__(self, operation_type):
		self.operation_type = operation_type
	def __repr__(self):
		return "%s@%s(%s)" % (self.typeName, self.values[self.host], gc.seqstr(self.values))

class IRule:
	#left is a predicate right is an operation
	def __init__(self,left,right):
		self.left = left
		self.right = right
	def setOriginalRuleNo(self,original_rule_no):#to support multiple execution plans
		self.original_rule_no = original_rule_no


class Index:
	def __init__(self, index):
		self.index = index
	def __repr__(self):
		return str(self.index)

class Constant:
	def __init__(self, constant):
		self.constant = constant
	def __repr__(self):
		return str(self.constant)+"_ct"

class IFunction:
	def __init__(self,func,param_indexes):
		self.func = func
		self.param_indexes = param_indexes
	def __repr__(self):
		rval = self.func.typeName + '('
		for p in self.param_indexes:
			rval += str(p)+ ','
		rval += ')'
		return rval

class IJoin(IOperation):
	#first and second are either operations or tables
	#attribute_indexes1 and 2 are two list of ids, "first" and "second" tables such that we join on
	#attr_idx[1] == attr_idx[2]
	def __init__(self, first, second, attribute_indexes1,attribute_indexes2,negation):
		self.operation_type = d.JOIN
		self.first = first
		self.second = second
		self.attribute_indexes1 = attribute_indexes1
		self.attribute_indexes2 = attribute_indexes2

		self.values = list(first.values)
		self.negation = negation
		self.typeName = "J [ "+first.typeName+" "+second.typeName+", {"
		for no,i in enumerate(attribute_indexes1):
			self.typeName += first.values[i.index].value+" "#+"("+str(i.index)+","+str(attribute_indexes2[no])+") "
		self.typeName += "} ]"
		if self.negation:
			self.typeName = '%s%s' % ('-',self.typeName)

		for vid,v in enumerate(second.values):
			#found = False
			#for id2 in attribute_indexes2:
			#	if vid == id2.index:
			#		found = True
			#if not found:
			#	self.values.append(v)
			#the previous are for deleting columns for joins
			self.values.append(v)
		#assume both are on the same host
		self.host = first.host
		self.type = gc.PredicateType.lookup(self)
	def __repr__(self):
		#smain = "%s@%s(%s)" % (self.typeName, self.values[self.host], gc.seqstr(self.values))
		#sfirst = self.first.__repr__()
		#sright = self.second.__repr__()
		#return "%s: first %s second %s" % (smain,sfirst,sright)
		return self.typeName
		
		
class IProject(IOperation):
	#attributes is a list of projected attributes or functions
	def __init__(self, table, attribute_indexes):
		self.operation_type = d.PROJECT
		self.table = table
		self.attribute_indexes = attribute_indexes
		self.values = table.values

		self.typeName = "P [ "+table.typeName+" , {"
		for idx in self.attribute_indexes:
			if isinstance(idx,Index):
				self.typeName += self.values[idx.index].__repr__()+" "
			if isinstance(idx,IFunction):
				self.typeName += idx.__repr__()+" "
		self.typeName += " } ]"
		
		self.host = table.host
		self.type = gc.PredicateType.lookup(self)
		
	def __repr__(self):
		#smain = "%s@%s(%s)" % (self.typeName, self.values[self.host], gc.seqstr(self.values))
		#stable = self.table.__repr__()
		#attributes = "{"
		#for idx in self.attribute_indexes:
		#	attributes += self.values[idx].__repr__()
		#attributes += "}"
		#return "%s (table %s) , %s " % (smain,stable,attributes)
		return self.typeName



class IAggregation(IOperation):
	def __init__(self, table, aggregation_type, attribute_index):
		self.operation_type = d.AGGREGATE_FUNC
		self.table = table
		self.aggregation_type = aggregation_type
		self.attribute_index = attribute_index

		self.typeName = "A [ "+table.typeName+" ]"
		self.values = table.values
		self.host = table.host
		self.type = gc.PredicateType.lookup(self)
	def __repr__(self):
		return self.typeName


class ISelect(IOperation):
	#the constructor takes as parameters three lists. the list of attributes, the operations on each of them
	# and the values on the other side of the operations
	def __init__(self, table, attribute_indexes, select_operators, right_values):
		self.operation_type = d.SELECT
		self.table = table
		self.attribute_indexes = attribute_indexes
		self.select_operators = select_operators
		self.right_values = right_values

		self.values = table.values
		self.typeName = "S [ "+table.typeName+" , {"
		for i,idx in enumerate(self.attribute_indexes):
			#self.typeName += self.values[idx.index].__repr__()+" "+comparison_oper_string[select_operators[i]]+" "+values[i]+","
			self.typeName += self.values[idx.index].__repr__()
			self.typeName += " "+self.select_operators[i]
			if(isinstance(self.right_values[i],Index)):
				self.typeName += " "+self.values[self.right_values[i].index].__repr__()+","
			else:
				self.typeName += " "+self.right_values[i].__repr__()+","
		self.typeName += " } ]"
		self.host = table.host
		self.type = gc.PredicateType.lookup(self)

	def __repr__(self):
		return self.typeName
