import defines as d
import grammar_classes as gc
from pdb import set_trace as st
import os
import sys

#classes in this file form a cleaner interface to be used by the nesC backend
#decoupled from what is before

TYPE_INT = 0x01
TYPE_UINT = 0x02
TYPE_CHAR = 0x04

SIZE_8 = 0x10
SIZE_16 = 0x20
SIZE_32 = 0x40

TYPE_INT8 =  TYPE_INT | SIZE_8
TYPE_UINT8 = TYPE_UINT | SIZE_8
TYPE_CHAR8 = TYPE_CHAR | SIZE_8

TYPE_INT16 = TYPE_INT | SIZE_16
TYPE_UINT16 = TYPE_UINT | SIZE_16
TYPE_CHAR16 = TYPE_CHAR | SIZE_16

TYPE_INT32 = TYPE_INT | SIZE_32
TYPE_UINT32 = TYPE_UINT | SIZE_32
TYPE_CHAR32 = TYPE_CHAR | SIZE_32



SELTYPE_EQUAL = 0
SELTYPE_NOTEQUAL = 1
SELTYPE_GREATER = 2
SELTYPE_GREATEROREQUAL = 3
SELTYPE_LESS = 4
SELTYPE_LESSOREQUAL = 5


def getTypeName(typeval):
	lookup = {TYPE_INT8:'int8_t',
			  TYPE_UINT8:'uint8_t',
			  TYPE_CHAR8:'char8_t',
			  TYPE_INT16:'int16_t',
			  TYPE_UINT16:'uint16_t',
			  TYPE_CHAR16:'char16_t',
			  TYPE_INT32:'int32_t',
			  TYPE_UINT32:'uint32_t',
			  TYPE_CHAR32:'char32_t',
			  }
	if typeval in lookup:
		return lookup[typeval]
	else:
		raise Exception, 'Invalid type num %s' % (typeval)


#classes that form the symbol table


class Column:
	#note that name is just for debug purposes and should not be relied upon
	def __init__(self, name, type, size):	
		self.name = name
		self.type = type
		self.size = size
		self.strtypename = 'unknown'
	def __init__(self, type, size):	
		self.name = "generic"
		self.type = type
		self.size = size
	def __repr__(self):
		return self.name

class SymbolTableEntry:
	def __init__(self, name):
		self.name = name
	def __repr__(self):
		return self.name

class TableFact:
	def __init__(self, symbol,values,compares,consts):
		self.symbol = symbol
		self.values = values
		self.compares = compares
		self.consts = consts

# this corresponds roughly to a grammar_class PredicateType, NOT Predicate
# translating from david's vocabulary to lucian's:
# [david] -> [lucian]
# builtin -> primary
# PredicateType -> Predicate
class Predicate(SymbolTableEntry):
	maxid = 0

	# todo: make this more general e.g. foo is hardcode
	name2id = {}
	try:
		f = open("foo")
		opened = True
	except IOError:
		opened = False
	if opened:
		for l in f:
			strs = l.split()
			pname = strs[0]
			pid = int(strs[1])
			name2id[pname] = pid

	#columns is a list of Column classes
	def __init__(self, name, primary, materialize, pkey, host, replacepk, priority, everTriggers):
		self.name = name
		self.temporary = False
		self.primary = primary
		self.materialize = materialize
		self.pkey = pkey
		self.columns = []
		self.host_idx = host
		self.replacepk = replacepk
		self.priority = priority
		self.assignid()
# 		self.assignid_old()
		#print "%s assigned %u" % (self.name, self.id)
		self.everTriggers = everTriggers
		#print '==================================================== '+name+' everTriggers '+str(everTriggers)
	def assignid(self):
		if self.name in Predicate.name2id:
			self.id = Predicate.name2id[self.name]
		else:
			while Predicate.maxid in Predicate.name2id.values():
				Predicate.maxid = Predicate.maxid + 1
			self.id = Predicate.maxid
			Predicate.maxid = Predicate.maxid + 1
	def assignid_old(self):
		self.id = Predicate.maxid
		Predicate.maxid = Predicate.maxid + 1
	def setTemporary(self):
		self.temporary = True
	def addColumn(self,column):
		self.columns.append(column)
	def __repr__(self):
		return self.name

class NescMaterialize:
	def __init__(self, timeout, cardinality, evictpolicy):
		self.timeout = timeout
		self.cardinality = cardinality
		self.evictpolicy = evictpolicy

class Function(SymbolTableEntry):
	maxid = 0
	def __init__(self, name, return_type, builtin):
		self.name = name
		self.parameters = []
		self.return_type = return_type
		self.builtin = builtin
		self.id = Function.maxid
		Function.maxid = Function.maxid + 1
		print "%s assigned %u" % (self.name, self.id)
	def addParameter(self,parameter):
		self.parameters.append(parameter)

class SymbolTable:
	def __init__(self):
		self.symbols = []
		
	def addSymbol(self,symbol):
		self.symbols.append(symbol)
	def lookupPredicateByName(self,predicateName):
		for s in self.symbols:
			if isinstance(s,Predicate) and s.name == predicateName:
				return s
		raise RuntimeError, 'predicate name not found "%s"' % (predicateName)
	def removePredicateByName(self,predicateName):
		for s in self.symbols:
			if isinstance(s,Predicate) and s.name == predicateName:
				self.symbols.remove(s)
				return True
		return False
	# todo: def __repr__(self):
		


#following are the operation classes
#it might have been nicer to have a base class called Operation but we're in Python here ...
#all parameters are pointers to the symbol table
#the implicit order of columns for result is table 1 columns then table 2 columns except those joined upon
globalopid = 0;
def getGlobalopid():
	global globalopid
	globalopid = globalopid + 1
	return globalopid

class EquiJoinOperation:
	maxid = 0
	def __init__(self, table_1, table_2, attributes_1,attributes_2,result,negation):
		self.table_1 = table_1
		self.table_2 = table_2
		self.attributes_1 = attributes_1
		self.attributes_2 = attributes_2
		self.result = result
		self.negation = negation
		self.nescname = 'JoinControl_IPS_ILS_OPS_M'
		self.id = EquiJoinOperation.maxid
		EquiJoinOperation.maxid = EquiJoinOperation.maxid + 1
		self.globalid = getGlobalopid()
	def __repr__(self):
		negstr = ''
		if self.negation:
			negstr = '-'
		strattributes = "("
		for i,a in enumerate(self.attributes_1):
			strattributes +="("+str(a)+","+str(self.attributes_2[i])+")"
		strattributes +=")"
		return negstr + "J ["+self.table_1.__repr__()+","+self.table_2.__repr__()+" on "+strattributes+"] -> "+self.result.__repr__()
	def gettables(self):
		return [self.table_1, self.table_2]	

class ProjectOperation:
	maxid = 0
	def __init__(self,table, attributes,result):
		self.table = table
		self.attributes = attributes
		self.result = result
		self.nescname = 'ProjectControl_IPS_OPS_M'
		self.id = ProjectOperation.maxid
		ProjectOperation.maxid = ProjectOperation.maxid + 1
		self.globalid = getGlobalopid()
	def __repr__(self):
		strattr = "("
		for a in self.attributes:
			strattr += str(a)+","
		strattr += ')'
		return "P ["+self.table.__repr__()+" on "+strattr +"] -> "+self.result.__repr__()
	def gettables(self):
		return [self.table]

class SelectOperation:
	maxid = 0
#the right_values can be either a constant or an index, see the intermediary_language classes
	def __init__(self, table, attributes, select_operators, right_values,result):
		self.table = table
		self.attributes = attributes
		self.select_operators = select_operators
		self.right_values = right_values
		self.result = result
		self.nescname = 'SelectControl_IPS_OPS_M'
		self.id = SelectOperation.maxid
		SelectOperation.maxid = SelectOperation.maxid + 1
		self.globalid = getGlobalopid() # todo should really have superclass
	def __repr__(self):
		strattr = "("
		for i,a in enumerate(self.attributes):
			strattr += str(a)+" "+self.select_operators[i]+" "+self.right_values[i].__repr__()+","
		strattr += ')'
		return "S ["+self.table.__repr__()+" on "+strattr+"] -> "+self.result.__repr__()
	def gettables(self):
		return [self.table]

class AggregationOperation:
	maxid = 0
	def __init__(self, table, aggregation_type, attribute_index,result):
		self.operation_type = d.AGGREGATE_FUNC
		self.table = table
		self.aggregation_type = aggregation_type
		self.attribute_index = attribute_index
		self.result = result
		self.nescname = 'AggregationControl_IPS_ILS_OPS_M'
		self.id = AggregationOperation.maxid
		AggregationOperation.maxid = AggregationOperation.maxid + 1
		self.globalid = getGlobalopid() # todo should really have superclass
	def __repr__(self):
		return "A [%s<%s of %s>] -> %s" % (self.aggregation_type,self.attribute_index,self.table,self.result)	
	def gettables(self):
		return [self.table]

#a rule class contains a list of operation classes

class Rule:
	def __init__(self, operations, broadcast_result):
		self.operations = operations
		self.broadcast_result = broadcast_result
	def addOperation(self,operation):
		self.operations.append(operation)
	def __repr__(self):
		res = ''
		for o in self.operations:
			res += o.__repr__()+'\n'
		return res
	def setOriginalRuleNo(self,original_rule_no):#to support multiple execution plans
		self.original_rule_no = original_rule_no
