#!/usr/bin/python

import os
import string
import sys
import templatelib as tl
import nesc_interface as nesC
import intermediary_language as il
import copy
import defines
from pdb import set_trace as st

#lookup types
JOIN_TYPE = 1
PROJECT_TYPE = 2
SELECT_TYPE = 3
AGGREGATION_TYPE = 4

STATIC_NO_TABLE_ROWS = 10
STATIC_MAX_CONNECTIONS = 3
STATIC_TIMEOUT_VALUE = 1000
TIMER_NO_TABLE_ROWS = STATIC_NO_TABLE_ROWS # TODO CLEARLY LEADS TO INEFFICIENT OVERALLOCATION AND INCORRECT UNDERALLOCATION!
WORD_SIZE = 2 # in bytes

param_store_component = 'ParamStoreM'
runtime_component = 'RuntimeComponentGeneratedM'
network_send_component_name = 'NetworkSendSyncM'
network_recv_component_name = 'NetworkReceiveSyncM'
uart_send_component_name = 'UartSendSyncM'
table2_cname = 'Table2_M'



ALL_TABLE_PARAMS_FILE = '_tableparams_gen_.c'

def getTypeId(operation):
	if isinstance(operation,nesC.EquiJoinOperation):
		return JOIN_TYPE
	if isinstance(operation,nesC.ProjectOperation):
		return PROJECT_TYPE
	if isinstance(operation,nesC.SelectOperation):
		return SELECT_TYPE
	if isinstance(operation,nesC.AggregationOperation):
		return AGGREGATION_TYPE


def primaryBasename(p):
	# Chop the 'C' if it is the last char
	name = p.primary
	if name[len(name)-1] == 'C':
		return name[0:(len(name)-1)]
	else:
		return name

def primaryComponentName(p):
	return "%sC" % (primaryBasename(p))

def primaryComponentFilename(p):
	return "%s.nc" % (primaryComponentName(p))

def primaryHeaderName(p):
	return "%s" % (primaryBasename(p))

def primaryHeaderFilename(p):
	return "%s.h" % (primaryHeaderName(p))

def primaryHeaderTemplateName(p):
	return "%s_template" % (primaryBasename(p))

def primaryHeaderTemplateFilename(p):
	return "%s.h" % (primaryHeaderTemplateName(p))
	
def find_template_file(p):
	return primaryHeaderTemplateFilename(p)
	
class NescWriter:
	def __init__(self, symboltable, plan, facts, templatefactory, out=None):
		self.symboltable = symboltable
		self.plan = plan
		self.factory = templatefactory
		self.out = out
		self.facts = facts
		if self.out != None and self.out != sys.stdout:
			pass # dcc todo: "rm -rf" is really what i need!			

	def openOutfile(self, filename, mode='w'):
		if self.out == None or self.out == sys.stdout:
			return sys.stdout
		else:
			extfilename = self.out + '/' + filename
			if not os.path.isdir(os.path.dirname(extfilename)):
				os.makedirs(os.path.dirname(extfilename))
			f = open(self.out + '/' + filename,mode)
			return f
			
	def getFactLocalTest(self,f,p):
		if str(f.values[0]) == defines.BCAST_ADDR :
			return 'TRUE'
		elif len(f.compares) == 0:
			return 'TOS_LOCAL_ADDRESS == '+str(f.values[p.host_idx].value)
		else:
			fact_test = ''
			for i,op in enumerate(f.compares):
				if i != 0:
					fact_test += ' && '
				fact_test += 'TOS_LOCAL_ADDRESS '+f.compares[i]+' '+f.consts[i]
			return copy.copy(fact_test)

	def writeTables(self):
		slowgo = 200;
		all_tableparams_decls = ''
		all_tableparams = {}
		all_facts_init = ''
		all_generated_debug_print_row = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				template_file = find_template_file(p)
			else:
				template_file = [ 'Table_decls_template.nc','Table_template.nc']
			
			#create the appropriate table
			hashmap = {'TableName':p.name + '_M'} # not very useful

			#get the actual values from DATALOG extensions
			hashmap['MAX_CONNECTIONS'] = str(STATIC_MAX_CONNECTIONS)
			hashmap['NUM_ROWS'] = p.materialize.cardinality
			hashmap['EVICT_POLICY'] = string.upper(p.materialize.evictpolicy)
			if p.materialize.timeout == 'infinite':
				hashmap['TIMEOUT_VALUE'] = 0
				hashmap['DO_TIMEOUT_TIMER'] = 'FALSE'				
			else:
				hashmap['TIMEOUT_VALUE'] = p.materialize.timeout
				hashmap['DO_TIMEOUT_TIMER'] = 'TRUE'
			

			hashmap['HOST_COLUMN'] = str(p.host_idx)
			if p.replacepk == True:
				hashmap['REPLACEPK'] = 'TRUE'
			else:
				hashmap['REPLACEPK'] = 'FALSE'
			hashmap['PRIORITY'] = p.priority
			hashmap['NUM_COLUMNS'] = str(len(p.columns))
			total_size = 0
			offset_array = ''
			for i,c in enumerate(p.columns):
				if i == 0:
					offset_array += '0'
				else:
					offset_array += ','+str(total_size)
				total_size += c.size
			row_padding = (-total_size % WORD_SIZE)
			hashmap['ROW_SIZE_IN_BYTES'] = total_size + row_padding
			hashmap['ROW_SIZE_IN_BYTES_UNPADDED'] = total_size
			hashmap['OFFSET_LIST'] = offset_array
			hashmap['OFFSET_LIST_SZ'] = len(p.columns)
			#primary key bits
			pk_bit_list = ''
			crtbitnum = 0;
			for i,c in enumerate(p.columns):
				if i%8 == 0 and not i == 0:
					pk_bit_list += str(crtbitnum)+','
					crtbitnum = 0;
				if p.pkey[i] == True:
					crtbitnum |= 1<<(i%8);
			if len(p.columns) > 8:
				pk_bit_list += ','
			pk_bit_list += str(crtbitnum);
			hashmap['PK_BIT_LIST'] = pk_bit_list
			hashmap['PK_BIT_LIST_SZ'] = str(len(p.columns)) + '/8+1'
			hashmap['MY_TYPE_ID'] = str(p.id)
			start_of_rule_table = False
			for tr in self.plan:
				if tr.operations[0].gettables()[0] == p:
					start_of_rule_table = True
					break
			if start_of_rule_table == True:
				hashmap['DO_LOCAL_NOTIFY'] = 'TRUE'
			else:
				hashmap['DO_LOCAL_NOTIFY'] = 'FALSE'
			if p.everTriggers == True:
				hashmap['EVER_TRIGGERS'] = 'TRUE'
			else:
				hashmap['EVER_TRIGGERS'] = 'FALSE'
			
# yyy need to support builtin's facts somehow,

			#support for timer facts
			# TODO "subclass" the standard table_template.
# 			if p.name == 'timer':
# 				fact_count = 0
# 				facts_init = ''
# 				for f in self.facts:
# 					if f.symbol == p:
# 						if fact_count >= TIMER_NO_TABLE_ROWS:
# 							raise RuntimeError, 'Probably Too many facts as compared to number of rows for table '+p.name+' TODO: OPTIMIZE CODE TO PUT FACTS SEQUENTIALLY'
# 						hash_fact ={}
# 						hash_fact['rownum'] = fact_count
# 						hash_fact['FACT_HOST'] = str(f.values[p.host_idx])
# 						hash_fact['LOCAL_FACT_TEST'] = self.getFactLocalTest(f,p)
# 						hash_fact['TIMER_ID'] = f.values[1]
# 						hash_fact['TIMER_VAL'] = f.values[2]
# 						fact_count += 1
# 						facts_init += self.factory.instantiate('timer_fact_initialization_template.nc',hash_fact)
# 				if fact_count != 0:
# 					hashmap['INITIALIZE_FACTS'] = facts_init
# 					hashmap['comment_facts_init'] = ''
# 				else:
# 					hashmap['INITIALIZE_FACTS'] = '//no facts'
# 					hashmap['comment_facts_init'] = '//no facts - '
# 			elif p.name == 'linklqi':
# 				fact_count = 0;
# 				for f in self.facts:
# 					if f.symbol == p:
# 						fact_count += 1 
# 				if fact_count != 0:
# 					raise RuntimeError, 'should implement facts for linklqi\n';
# 				else:
# 					hashmap['INITIALIZE_FACTS'] = '//no facts'
# 					hashmap['comment_facts_init'] = '//no facts - '
					
			if p.primary:
				pass
			#support for all other facts
			else:
				fact_count = 0
				factsperhost = {}
				facts_init = ''
				for f in self.facts:
					if f.symbol == p:
						hostaddr = f.values[p.host_idx].value
						if hostaddr not in factsperhost:
							factsperhost[hostaddr] = 0
						if factsperhost[hostaddr] >= int(p.materialize.cardinality):
							raise RuntimeError, 'Probably Too many facts as compared to number of rows for table='+p.name+', hosstaddr='+hostaddr
						hash_fact ={}
						hash_fact['MY_TYPE_ID'] = p.id
						hash_fact['ROWNUM'] = factsperhost[hostaddr]
						hash_fact['FACT_HOST'] = str(hostaddr)
						general_fact = True
						#first is a special test for all
						hash_fact['LOCAL_FACT_TEST'] = self.getFactLocalTest(f,p)
						column_init_list = ''
						for cno,c in enumerate(p.columns):
							if cno == p.host_idx: # we could add: and general_fact == True:
								colvalue = 'TOS_LOCAL_ADDRESS'
							else:
								colvalue = str(f.values[cno]);
							column_init_list += '\n\t\t\t*(('+c.strtypename+'*)(row+((p->offset_array)['+str(cno)+']))) = '+colvalue+';'
						hash_fact['COLUMN_INITLIALIZATION_LIST'] = column_init_list
						fact_count += 1
						factsperhost[hostaddr] += 1
						facts_init += self.factory.instantiate('fact_initialization_template.nc',hash_fact)

				if fact_count != 0:
					hashmap['DO_INITIAL_FACTS_NOTIFY'] = 'TRUE'
				else:
					hashmap['DO_INITIAL_FACTS_NOTIFY'] = 'FALSE';
				all_facts_init += '\t\t// Facts for %s\n' % (p.name)
				all_facts_init += '\t\tp = &(_tableparams_[%u]);\n' % (p.id)
				all_facts_init += '\t\tcall FactKicker.start[%u](TIMER_ONE_SHOT,%u);\n' % (p.id,slowgo)
				all_facts_init += facts_init

				slowgo += 100;

			#print row debug for TAM by dav copying lucian
			# dcc todo: should eventually consolidate to one of the print stmts
			generated_debug_print_row = '//GeneratedTamRowPrint\n'
			for i,c in enumerate(p.columns):
# 				generated_debug_print_row += '\t\tprintf("(col=%u,vt=%u,val=%s,hex=0x%s)",*(%s*)(tdata+p->offset_array[%u]),*(%s*)(tdata+p->offset_array[%u]));\n' % (i,c.type,'%d','%x',c.strtypename,i,c.strtypename,i)
				generated_debug_print_row += '\t\tprintf("(col=%u,val=%s,hex=0x%s)",*(%s*)(tdata+p->offset_array[%u]),*(%s*)(tdata+p->offset_array[%u]));\n' % (i,'%d','%x',c.strtypename,i,c.strtypename,i)
				if i != len(p.columns) - 1:
					generated_debug_print_row += '\t\tprintf(",");\n'
			all_generated_debug_print_row += 'case (%u):\n' % p.id
			all_generated_debug_print_row += generated_debug_print_row
			all_generated_debug_print_row += 'break;\n'
# 			hashmap['GeneratedTamRowPrint'] = generated_debug_print_row

			output = self.factory.instantiate(template_file, hashmap)
			if p.primary:
# 				f = self.openOutfile('tables/' + p.name+'_M.nc')
				f = self.openOutfile('tables/' + primaryHeaderFilename(p))
				print >>f,output
			else:
				all_tableparams_decls += output[0] # array decls
				all_tableparams[p.id] = output[1] # struct decls

		# all non-primary tables write to the same param file
		if len(all_tableparams.keys()) == 0:
			maxtt = 0
		else:
			maxtt = max(all_tableparams.keys()) + 1
		str_all_tableparams = ''
		for key in range(maxtt):
			if key in all_tableparams:
				str_all_tableparams += all_tableparams[key]
			else:
				str_all_tableparams += '{.tt = TUPLETYPE_NULL}, // %u is not autogen \n' % (key)

		f = self.openOutfile(ALL_TABLE_PARAMS_FILE)
		print >>f, """
%s


// these should be equal!: \"TTMAX\" and size of array below

TableParams_t _tableparams_[%u] = {
%s
};

#define LOOKUP(tt) (&(_tableparams_[(tt)]))
#define RLOOKUP(p) ((((void*)(p)) - (void*)_tableparams_)/sizeof(TableParams_t))

		""" % (all_tableparams_decls, maxtt, str_all_tableparams)

		f2 = self.openOutfile('_MAX_TUPLE_TYPE.h')
		print >>f2,'#define MAX_TUPLE_TYPE '+ str(maxtt)
		
		hashmap = {}
		hashmap['TABLE_NAME'] = table2_cname
		hashmap['ROW_PRINT'] = all_generated_debug_print_row
		hashmap['INITIALIZE_FACTS'] = all_facts_init
		output = self.factory.instantiate('Table2_template.nc', hashmap)
		f = self.openOutfile(table2_cname+'.nc')
		print >>f,output

	def writeWiring(self):
		# Major TODO note: When we move to async interfaces,
		# the wiring done here will be dangerous if two
		# components A,B use the same component C.  The reason
		# is: C's signals will propagate to both A and B.
		# Then we will have to think of another solution...

		hashmap = {'ConfigurationName':'GlobalConfiguration'}

		#stdcontrol wirings
		std_wiring = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.temporary:
				continue
			if p.primary:
				std_wiring += primaryComponentName(p)+'.StdControl'+'<-'+runtime_component+'.LowerControl;\n\t'
		hashmap['StdControlWiring'] = std_wiring
		
		
		#notify wirings
		notify_wiring = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				notify_wiring += primaryComponentName(p)+'.NotifyFresh'+'<-'+runtime_component+'.NotifyFresh['+str(p.id)+'];'+'\n\t'
			else:
				notify_wiring += table2_cname+'.NotifyFresh['+str(p.id)+']<-'+runtime_component+'.NotifyFresh['+str(p.id)+'];'+'\n\t'
		hashmap['NotifyWiring'] = notify_wiring

		# Timer wirings
		factkicker_wiring = ''
		garbage_collector_wiring = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				continue
			factkicker_wiring += '%s.FactKicker[%u]->TimerC.Timer[unique("Timer")];\n\t' % (table2_cname, p.id)
			garbage_collector_wiring += '%s.GarbageTupleCollector[%u]->TimerC.Timer[unique("Timer")];\n\t' % (table2_cname, p.id)
		hashmap['TimerFactKickerWiring'] = factkicker_wiring
		hashmap['TimerGarbageCollectorWiring'] = garbage_collector_wiring


		#generic function call wiring		
		func_wiring = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				func_wiring += primaryComponentName(p)+'.GenericFunctionCall->GenericFunctionCallC.GenericFunctionCall;\n\t'
			else:
				pass # only one
# 				func_wiring += '%s.GenericFunctionCall[%u]->GenericFunctionCallM.GenericFunctionCall;\n\t' % (table2_cname, p.id)
		hashmap['FunctionWiring'] = func_wiring

		
		
		#Enumerate operations (nesC components)
		namelist = []
		countlist = []
		atleastonejoin = False
		atleastoneagg = False
		atleastonesel = False
		for r in self.plan:
			for o in r.operations:
				if not o.nescname in namelist:
					namelist.append(o.nescname);
		for n in namelist:
			count = 0
			for r in self.plan:
				for o in r.operations:
					if o.nescname == n:
						if isinstance(o,nesC.EquiJoinOperation):
							atleastonejoin = True
						if isinstance(o,nesC.AggregationOperation):
							atleastoneagg = True
						if isinstance(o,nesC.SelectOperation):
							atleastonesel = True
						count += 1
			countlist.append(count);

		# adding main components
		component_list = [] # dcc todo: this should really only add tables if they are actually used in rules or facts, not if they are simply in typedef or builtin statements...
		for i,n in enumerate(namelist):
			cname = 'new %s(%u)' % (n,countlist[i])
			component_list.append(cname)

		# add the extra components
		component_list.append(param_store_component)
		nonprimarymaxid = -1
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary:
				continue
			if p.primary:
				cname = '%sC' % (p.primary)
				component_list.append(cname)
			elif p.id > nonprimarymaxid:
				nonprimarymaxid = p.id # this is actually less efficient but better for having table2 TTMAX == actual number of tables than another layer of mapping
		atleastonenonprimary = False
		if nonprimarymaxid >= 0:
			atleastonenonprimary = True
			cname = 'new %s(%u)' % (table2_cname, nonprimarymaxid+1)
			component_list.append(cname)
		hashmap['ComponentList'] = ','.join(component_list)

				
		error_exit_link = '';
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary:
				continue
			if p.primary:
				continue # alot of the primary guys don't have ErrorExit implemented
# 				error_exit_link += p.name+ '_M.ErrorExit -> ErrorExitM;\n'

		if nonprimarymaxid >= 0:
			error_exit_link += 'Table2_M.ErrorExit -> ErrorExitM;\n'
		hashmap['ErrorExitLink'] = error_exit_link;

# yyy: custom wirings for builtins
# 		if atleasttimerm:
# 			timer_component_list = ''
# 			for i in range(TIMER_NO_TABLE_ROWS):
# 				timer_component_list += 'timer_M.Timer[%u]->TimerC.Timer[unique("Timer")];\n' % (i)
# 			hashmap['AtLeastTimerM'] = timer_component_list
# 		else:
# 			hashmap['AtLeastTimerM'] = '// No timer_M'				

		
		#wiriting to the gam (including tam)
		gam_links = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				cname = primaryComponentName(p)
				gam_links += 'GtupleAccessorMutatorM.Tam['+str(p.id)+']'+'->'+cname+'.Tam;\n\t'
				gam_links += cname+'.Gam'+'->'+'GtupleAccessorMutatorM.Gam;\n\t'
			else:
				gam_links += 'GtupleAccessorMutatorM.Tam['+str(p.id)+']'+'->'+table2_cname+'.Tam['+str(p.id)+'];\n\t'
		hashmap['GamLinks'] = gam_links		
		
		#wiring for the chain of operations
		chain_wiring = ''
		for r in self.plan:
			chain_wiring += '\t//Rule '+str(r.original_rule_no)+'\n'
			for o1 in r.operations:			
					for o2 in r.operations:
						if o1.result in o2.gettables() and o1.result.temporary:
							chain_wiring += '\t'+self.writewiringop2op(o1,o2)+'\n'
							
		hashmap['ChainWiring'] = chain_wiring
		
		# standard wiring for joins
		if atleastonejoin:
			hashmap['AtLeastOneJoin'] = 'JoinControl_IPS_ILS_OPS_M.Join->JoinM.Join;\n'# todo should use nescname in nesc_interface
			hashmap['AtLeastOneJoin'] += 'JoinControl_IPS_ILS_OPS_M.Gam->GtupleAccessorMutatorM.Gam;\n'
		else:
			hashmap['AtLeastOneJoin'] = '// No Joins'

		# standard wirings for aggregations
		if atleastoneagg:
			aggconfig = 'AggregationControl_IPS_ILS_OPS_M.Gam->GtupleAccessorMutatorM.Gam;\n'
			import defines as d
			import string
			for i,aggname in enumerate(d.AGGREGATION_TYPES):
				aggcomponent = 'Aggregation%sC'%(string.capitalize(aggname))
				aggconfig += 'components %s;\n' % (aggcomponent)
				aggconfig += 'AggregationControl_IPS_ILS_OPS_M.Aggregation[%u]->%s.Aggregation;\n' % (i,aggcomponent) # todo should use nescname in nesc_interface
			hashmap['AtLeastOneAgg'] = aggconfig
		else:
			hashmap['AtLeastOneAgg'] = '// No Aggregations'

		if atleastonesel:
			selconfig = 'SelectControl_IPS_OPS_M.Selection->SelectionM;\n'
			hashmap['AtLeastOneSel'] = selconfig
		else:
			hashmap['AtLeastOneSel'] = '// No Selection'
			pass
		
		#wiring operations with tables
		opnum2tt = {}
		operation_table_wiring = ''
		for r in self.plan:
			for o in r.operations:
				if isinstance(o,nesC.EquiJoinOperation):
					#nasty hack if both one is temporary pick the other else pick the second one
					pulltable = None
					if o.table_2.temporary:
						pulltable = o.table_1
					else:
						pulltable = o.table_2
					if pulltable.primary:
						pullname = primaryComponentName(pulltable)
						pulliface = 'LocalPull'
						operation_table_wiring += o.nescname+'.In2_Pull['+str(o.id)+']'+ '->'+pullname+'.'+pulliface+';\n\t' # dcc todo: danger in2 assump!
					else:
						pulliface = 'LocalPull'
						operation_table_wiring += o.nescname+'.In2_Pull['+str(o.id)+']'+ '->'+table2_cname+'.'+pulliface+'['+str(pulltable.id)+'];\n\t' # dcc todo: danger in2 assump!
				if isinstance(o,nesC.ProjectOperation):
					if o.result.primary:
						endpoint = primaryComponentName(o.result)+'.LocalPush['+str(o.id)+']'
					else:
						endpoint = table2_cname+'.LocalPush['+str(o.id)+']' # correct?
						opnum2tt[o.id] = o.result.id
					operation_table_wiring += o.nescname+'.Out_Push['+str(o.id)+']'+'->'+endpoint+';\n\t'
				if isinstance(o,nesC.AggregationOperation):
					if o.table.primary:
						operation_table_wiring += o.nescname+'.In2_Pull['+str(o.id)+']'+ '->'+primaryComponentName(o.table)+'.LocalPull'+';\n\t'
					else:
						operation_table_wiring += o.nescname+'.In2_Pull['+str(o.id)+']'+ '->'+table2_cname+'.LocalPull['+str(o.table.id)+'];\n\t'
# 					operation_table_wiring += o.nescname+'.In2_Pull['+str(o.id)+']'+ '->'+o.table.name+'_M.LocalPull'+';\n\t'
					# dcc TODO: double wiring is incorrect because the second wiring overwrites the results of first wiring when any function is called or is event is signaled.  the current wiring scheme produces duplicates between open/close which is not correct nor ideal.
# 					operation_table_wiring += o.nescname+'.In2_Pull['+str(o.id)+']'+ '->'+o.table.name+'_M.FreshPull'+';\n\t'

# 				operation_table_wiring += o.nescname+'.Join->JoinM.Join;\n\t'
		hashmap['ComponentTableWiring'] = operation_table_wiring
		
		print 'Opnum To TupleType mapping:', opnum2tt
		stropnum2tt = ''
		if len(opnum2tt.keys()) == 0:
			maxopnum = 0
		else:
			maxopnum = max(opnum2tt.keys()) + 1
		for opnum in range(maxopnum):
			if opnum in opnum2tt:
				stropnum2tt += '%u,' % (opnum2tt[opnum])
			else:
				stropnum2tt += 'TUPLETYPE_NULL,' # something to trigger error
		f = self.openOutfile('_opnum2tt_gen_.c')
		print >>f, """
TupleType_t _opnum2tt_[%u] = {%s};
#define LOOKUP_BY_OPNUM(opnum) (_opnum2tt_[opnum])
		""" % (maxopnum, stropnum2tt)
		
		#main component wiring
		main_task_wiring = ''
		#pulls from tables
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary
				continue
			if p.primary:
				main_task_wiring += runtime_component+'.PullInFresh['+str(p.id)+']'+'->'+primaryComponentName(p)+'.FreshPull'+';\n\t'
			else:
				main_task_wiring += runtime_component+'.PullInFresh['+str(p.id)+']'+'->'+table2_cname+'.FreshPull['+str(p.id)+'];\n\t'

		#pushes to operators
		for r in self.plan:
			o = r.operations[0]
# 			main_task_wiring += runtime_component+'.PushOut['+str(o.id)+']'+'->'+o.nescname+'.In1_Push['+str(o.id)+']'+';\n\t'
			main_task_wiring += runtime_component+'.PushOut['+str(o.globalid)+']'+'->'+o.nescname+'.In1_Push['+str(o.id)+']'+';\n\t'
		hashmap['MainTaskWiring'] = main_task_wiring
		
		
		
		#network wiring
		network_wiring = ''
		uart_wiring = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				cname = primaryComponentName(p)
				network_wiring += network_send_component_name+'.PullInSendable['+str(p.id)+']'+'->'+cname+'.RadioPull'+';\n\t'
				network_wiring += network_send_component_name+'.NotifyRadio['+str(p.id)+']'+'->'+cname+'.NotifyRadio'+';\n\t'
				network_wiring += network_send_component_name+'.Tam['+str(p.id)+']'+'->'+cname+'.Tam'+';\n\t'
				# uart +, this should later disappear with PumpM abstraction
				uart_wiring += uart_send_component_name+'.PullIn['+str(p.id)+']'+'->'+cname+'.UartPull'+';\n\t'
				uart_wiring += uart_send_component_name+'.Notify['+str(p.id)+']'+'->'+cname+'.NotifyUart'+';\n\t'
				uart_wiring += uart_send_component_name+'.Tam['+str(p.id)+']'+'->'+cname+'.Tam'+';\n\t'
			else:
				network_wiring += network_send_component_name+'.PullInSendable['+str(p.id)+']'+'->'+table2_cname+'.RadioPull['+str(p.id)+'];\n\t'
				network_wiring += network_send_component_name+'.NotifyRadio['+str(p.id)+']'+'->'+table2_cname+'.NotifyRadio['+str(p.id)+'];\n\t'
				network_wiring += network_send_component_name+'.Tam['+str(p.id)+']'+'->'+table2_cname+'.Tam['+str(p.id)+'];\n\t'
				# uart +, this should later disappear with PumpM abstraction
				uart_wiring += uart_send_component_name+'.PullIn['+str(p.id)+']'+'->'+table2_cname+'.UartPull['+str(p.id)+'];\n\t'
				uart_wiring += uart_send_component_name+'.Notify['+str(p.id)+']'+'->'+table2_cname+'.NotifyUart['+str(p.id)+'];\n\t'
				uart_wiring += uart_send_component_name+'.Tam['+str(p.id)+']'+'->'+table2_cname+'.Tam['+str(p.id)+'];\n\t'
				# uart -
				network_wiring += network_recv_component_name+'.PushOut['+str(p.id)+']'+'->'+table2_cname+'.RadioPush['+str(p.id)+'];\n\t'
		hashmap['NetworkWiring'] = network_wiring
		hashmap['UartWiring'] = uart_wiring


		
		#wiring the ParamStore
		#the previous version
		namecopy = copy.copy(namelist)
		param_wiring = ''
		for r in self.plan:
			for o in r.operations:
				if o.nescname in namecopy:
					namecopy.remove(o.nescname)
					typeid = getTypeId(o)
					param_wiring += o.nescname+'.ParamStore->'+param_store_component+'.ParamStore['+str(typeid)+'];\n\t'
		for p in self.symboltable.symbols:
			if isinstance(p,nesC.Predicate) and not p.temporary:
				if p.primary:
					param_wiring += primaryComponentName(p)+'.getParams->'+param_store_component+'.ParamStore['+str(PROJECT_TYPE)+'];\n\t'
				else:
					pass
					# skip for nonprimary - only one wiring
# 					param_wiring += table2_cname'.getParams['+str(p.id)']->'+param_store_component+'.ParamStore['+str(PROJECT_TYPE)+'];\n\t'
		hashmap['ParamStoreWiring'] = param_wiring


		# LEDs wirings
		led_wiring = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			if p.primary:
				led_wiring += primaryComponentName(p)+'.Leds -> LedsC'+';\n\t'
			else:
				# only one wiring
				pass
		for n in namelist:
			led_wiring += n+'.Leds -> LedsC'+';\n\t'
		hashmap['LedWiring'] = led_wiring

		#writing the output
		output = self.factory.instantiate('Config_template.nc', hashmap)
		
		f = self.openOutfile('GlobalConfiguration.nc')
		print >>f,output

	def writewiringop2op(self, inp, out):
		return inp.nescname+'.Out_Push['+str(inp.id)+']'+ '->'+out.nescname+'.In1_Push['+str(out.id)+'];'
	
		
	maxuid = 0
	@staticmethod
	def uid():
		NescWriter.maxuid = NescWriter.maxuid + 1
		return NescWriter.maxuid

	def writeParams(self):
		self.writeRulesParams()
		pass

	JOIN_INIT = ''
	SELECT_INIT = ''
	PROJECT_INIT = ''
	AGG_INIT = ''
	def writeRulesParams(self):
		hashmap = {}
		hashmap['JOIN_NO'] = nesC.EquiJoinOperation.maxid
		hashmap['PROJECT_NO'] = nesC.ProjectOperation.maxid
		hashmap['SELECT_NO'] = nesC.SelectOperation.maxid
		hashmap['AGGREGATE_NO'] = nesC.AggregationOperation.maxid

		generated_structures = ''
		initializations = ''		
		for r in self.plan:
			generated_structures += '// ------------------------- NEW RULE\n'
			initializations += '// ------------------------- NEW RULE\n'
			(tmp_gen_struct,tmp_init) = self.writeRuleParams(r)
			generated_structures += tmp_gen_struct
			initializations += tmp_init
		hashmap['GENERATED_STRUCTURES'] = generated_structures
		hashmap['PARAM_INITIALIZATION'] = initializations
		hashmap['JOIN_INIT'] = NescWriter.JOIN_INIT
		hashmap['SELECT_INIT'] = NescWriter.SELECT_INIT
		hashmap['PROJECT_INIT'] = NescWriter.PROJECT_INIT
		hashmap['AGG_INIT'] = NescWriter.AGG_INIT
		output = self.factory.instantiate('ParamStore_template.nc', hashmap)
		f = self.openOutfile('ParamStoreM.nc')
		print >>f,output
		print 'DONE +++++++++++++++++++++'
		
	def writeRuleParams(self,rule):
		hashmap = {}
		# rule's gtuple (passed horizontally)
		
		structure_gen = ''
		initializations = ''
		rulegtuid = NescWriter.uid()

		opcount = 1
		# per operator params
		for o in rule.operations:
			param_hashmap = {}
			paramuid = NescWriter.uid()
			structure_gen += '// --------- NEW Operator %s scope start' % (paramuid)+' \n'
			type_string = 'unknown'
			
			###################################
			#new join params
			if isinstance(o,nesC.EquiJoinOperation):
				# equijoin inparams
				type_string = 'join'
				hashmap = {}
				hashmap['UID'] = str(paramuid)
				structure_gen +=  '// ----- JoinParams' +str(o.id)+' (global uid '+str(paramuid)+ ') scope start' +'\n'
				hashmap['MAX_SIZE'] = len(o.attributes_1)
				hashmap['INSIZE'] = opcount
				if o.negation:
					isnegation = 'TRUE'
				else:
					isnegation = 'FALSE'
				opcount += 1 # dcc: Assume that pull side is size 1 e.g. left-deep joining! 
				structure_gen += 'ColnumPairTyped_t cptlist'+str(paramuid)+'['+str(hashmap['MAX_SIZE'])+ '] = {'
				cptlist = ''
				for i in range(len(o.attributes_1)):
					cptlist = '{'+str(o.attributes_1[i])+','+ str(o.attributes_2[i])+','+str(o.table_2.columns[o.attributes_2[i].index].type)+'}'
					if i != 0:
						structure_gen += ','+cptlist
					else:
						structure_gen += cptlist
				structure_gen += '};\n'

				structure_gen += 'Tuple_t t_inparam'+str(paramuid)+';\n'
				structure_gen += 'Gtuple_t gt_inparam'+str(paramuid)+'={&t_inparam'+str(paramuid)+',0,1};\n'
				structure_gen += 'JoinParams_t inparam%u={cptlist%u,%u,&gt_inparam%u,{%u},%s};\n' % (paramuid,paramuid,hashmap['MAX_SIZE'],paramuid,hashmap['INSIZE'],isnegation)

			###################################
			#projection params
			if isinstance(o,nesC.ProjectOperation):
				type_string = 'project'
				structure_gen += '// ----- ProjParams %s scope start' % (paramuid)+'\n'
				hasfunction = False;
				for i in range(len(o.attributes)):
					if not isinstance(o.attributes[i],il.Index):
						hasfunction = True;
						break;
				if not hasfunction:
					built_init = ''
					for i in range(len(o.attributes)):
						attr = o.attributes[i]
						varname = 'col_toget%u_%u' % (o.id,i)
						structure_gen += 'uint8_t %s = %u;\n' % (varname, attr.index)
						if i != 0:
							built_init += ','
						built_init += '{ EXPR_COL, &'+varname+',{'+str(i)+','+str(o.result.columns[i].type)+'}'+'}'
					built_init += ''
					structure_gen += 'ExprColnumTyped_t ectlist'+str(o.id)+'['+str(len(o.attributes))+'] = {'+built_init+'};\n'
				else:
					ectlist = ''
					listinit = ''
					for i in range(len(o.attributes)):
						attr = o.attributes[i]
						ectuid = NescWriter.uid()
						ectlist = ectlist + 'ect' + str(ectuid) + ','
						(defines,ini) = self.writeExprColnum(ectuid, attr, i, o)
						structure_gen += defines
						if i <> 0:
							listinit += ','
						listinit += ini
					structure_gen += 'ExprColnumTyped_t ectlist'+str(o.id)+'['+str(len(o.attributes))+'] = { '+listinit+' };\n'
				structure_gen += 'ProjParams_t inparam'+str(paramuid)+' ={&ectlist'+str(o.id)+'[0],'+str(len(o.attributes))+'};\n'
				
			###################################
			#aggregation
			if isinstance(o,nesC.AggregationOperation):
				type_string = 'aggregation'
				hashmap = {}
				structure_gen += '// ---- AggParams_t %s scope start\n' % (paramuid)
				structure_gen += 'Tuple_t inparam%u_gt_tlist[1] = {{TUPLETYPE_NULL, NULL}};\n' % (paramuid)
				structure_gen += 'Gtuple_t inparam'+str(paramuid)+'_gt = {inparam'+str(paramuid)+'_gt_tlist,0,1};\n'
				hashmap['UID'] = paramuid
				hashmap['COLIDX'] = o.attribute_index
				hashmap['COLTYPE'] = o.table.columns[o.attribute_index].type
				import defines as d
				if o.aggregation_type in d.AGGREGATION_TYPES:
					aggtypeid = d.AGGREGATION_TYPES.index(o.aggregation_type)
				else:
					raise RuntimeError, 'Aggregation %s invalid' % (o.aggregation_type)
				hashmap['AGGTYPE'] = aggtypeid
				hashmap['INSIZE'] = opcount

				(tmp_defines,tmp_instrs) = self.writeAggState(paramuid, aggtypeid)
				structure_gen += tmp_defines
				initializations += tmp_instrs

				structure_gen += 'AggParams_t inparam'+str(paramuid)+' = {'+str(hashmap['AGGTYPE'])+',{'+str(hashmap['COLIDX'])+','+str(hashmap['COLTYPE'])+'},&inparam'+str(paramuid)+'_aggstate,&inparam'+str(paramuid)+'_gt,{'+str(hashmap['INSIZE'])+'}};\n'

			###################################
			#selection
			if isinstance(o,nesC.SelectOperation):
				type_string='select'
				# DCC MAJOR TODO: THIS IS NOT GENERAL AT ALL AND PRETTY MUCH SUPPORTS ONLY COLUMN TO COLUMN COMPARISONS, MUCH LESS GENERAL THAN WHAT THE DATA STRUCTURES ARE DESIGNED FOR.  SELPARAMS_TEMPLATE.NC ALSO NEEDS MAJOR FIXING
				for i,lhs in enumerate(o.attributes):
					# TODO: should really make lhs and rhs functions
					paramuid = NescWriter.uid()
					hashmap = {}
					structure_gen += '// ---- SelParams_t %s scope start\n' % (paramuid)
					initializations += '// ----- SelParams_t %s scope start\n' %(paramuid)

					hashmap['UID'] = paramuid

					# lhs block
					if isinstance(lhs,il.Index):
						hashmap['LHS_TYPE'] = 'EXPR_COL'
						varname_suffix = '%u_lhs' % (paramuid)
						varname = 'colnum%s' % (varname_suffix)
						hashmap['LHS_VARNAME'] = varname
						
						colnumhashmap = {}
						colnumhashmap['UID'] = varname_suffix
						colnumhashmap['COL_NUM'] = lhs.index
						colnumhashmap['COL_TYPE'] = o.table.columns[lhs.index].type #nesC.TYPE_UINT16
						structure_gen += 'ColnumTyped_t '+varname+' = {'+ str(colnumhashmap['COL_NUM'])+','+str(colnumhashmap['COL_TYPE']) +'};\n'
					elif isinstance(lhs,il.Constant):
						raise NotImplementedError
					elif isinstance(lhs,il.IFunction):
						raise NotImplementedError
					else:
						raise Exception, 'SelectOperation is taking invalid type'

					# rhs block
					rhs = o.right_values[i]
					if isinstance(rhs,il.Index):
						hashmap['RHS_TYPE'] = 'EXPR_COL'
						varname_suffix = '%u_rhs' % (paramuid)
						varname = 'colnum%s' % (varname_suffix)
						hashmap['RHS_VARNAME'] = varname
						colnumhashmap = {}
						colnumhashmap['UID'] = varname_suffix
						colnumhashmap['COL_NUM'] = rhs.index
						colnumhashmap['COL_TYPE'] = o.table.columns[rhs.index].type
						structure_gen += 'ColnumTyped_t '+varname+' = {'+str(colnumhashmap['COL_NUM'])+','+str(colnumhashmap['COL_TYPE'])+'};\n'
					elif isinstance(rhs,il.Constant):
						hashmap['RHS_TYPE'] = 'EXPR_VAL'
						varname_suffix = '%u_rhs' % (paramuid)
						varname = 'val%s' % (varname_suffix)
						hashmap['RHS_VARNAME'] = varname
						valhashmap = {}
						valhashmap['UID'] = varname_suffix
						actualname = '%s_actual' % (varname)
						valhashmap['VARNAME'] = actualname
						actualtype = nesC.getTypeName(o.table.columns[lhs.index].type)
# 						pdb.set_trace()
						structure_gen += '%s %s = %s;\n' % (actualtype, actualname, rhs.constant)
						valhashmap['TYPE'] = o.table.columns[lhs.index].type
						structure_gen += 'ValTyped_t '+varname+' = {&'+str(valhashmap['VARNAME'])+','+str(valhashmap['TYPE'])+'};\n'
					elif isinstance(rhs,il.IFunction):
						raise NotImplementedError
					else:
						raise Exception, 'SelectOperation is taking invalid type'

					# select operation
					selop = o.select_operators[i]
					if selop == '==':
						hashmap['TYPE'] = nesC.SELTYPE_EQUAL
					elif selop == '!=':
						hashmap['TYPE'] = nesC.SELTYPE_NOTEQUAL
					elif selop == '>':
						hashmap['TYPE'] = nesC.SELTYPE_GREATER
					elif selop == '>=':
						hashmap['TYPE'] = nesC.SELTYPE_GREATEROREQUAL
					elif selop == '<':
						hashmap['TYPE'] = nesC.SELTYPE_LESS
					elif selop == '<=':
						hashmap['TYPE'] = nesC.SELTYPE_LESSOREQUAL
					else:
						raise Exception, 'Comparison not supported!'

					structure_gen += 'SelParams_t inparam'+str(paramuid)+'= {'+str(hashmap['TYPE'])+',{'+str(hashmap['LHS_TYPE'])+',&'+hashmap['LHS_VARNAME']+'},{'+str(hashmap['RHS_TYPE'])+',&'+hashmap['RHS_VARNAME']+'}};\n'
					structure_gen += '// ---- SelParams_t %s scope end\n' % (paramuid)
					initializations += '// ----- SelParams_t %s scope end\n' %(paramuid)
			
			param_hashmap['UID'] = str(paramuid)
			param_hashmap['type'] = type_string
			param_hashmap['id_in_type'] = o.id
			param_hashmap['GT_UID'] = str(rulegtuid)
			if type_string == 'join':
				if NescWriter.JOIN_INIT != '':
					NescWriter.JOIN_INIT += ','
				NescWriter.JOIN_INIT += ' {&inparam'+str(paramuid)+'} '
			if type_string == 'select':
				if NescWriter.SELECT_INIT != '':
					NescWriter.SELECT_INIT += ','
				NescWriter.SELECT_INIT += ' {&inparam'+str(paramuid)+'} '
			if type_string == 'project':
				if NescWriter.PROJECT_INIT != '':
					NescWriter.PROJECT_INIT += ','
				NescWriter.PROJECT_INIT += ' {&inparam'+str(paramuid)+'} '
			if type_string == 'aggregation':
				if NescWriter.AGG_INIT != '':
					NescWriter.AGG_INIT += ','
				NescWriter.AGG_INIT += ' {&inparam'+str(paramuid)+'} '

		structure_gen += '\n'
		initializations += '\n'
		return (structure_gen,initializations)
				
	def writeAggState(self,paramuid,aggtypeid):
		# dcc todo: aggstate is terrible and wrong - should be rewritten to take group by from the left hand side, not right hand side!
		aggtlistlen = 20 # dcc TODO ugly figure this out from prev chain
		extraallocsize = 30 # dcc todo so wasteful...
		tmp_defines = ''
		tmp_instrs = ''
		import defines as d
		aggname = string.capitalize(d.AGGREGATION_TYPES[aggtypeid])
		if aggname == 'Min':
			innertuplename = 'inparam%u_aggstate_min_gtuple_tlist' % (paramuid)
		elif aggname == 'Avg':
			innertuplename = 'inparam%u_aggstate_avg_gtuple_tlist' % (paramuid)
		else:
			raise RuntimeError, 'Aggregations other than Min need to carefully follow Min template standard with tlist or revise code below'
		hashmap = {}
		hashmap['UID'] = paramuid
		hashmap['GT_MAX'] = aggtlistlen
		
		tmp_defines2 = 'Tuple_t %s[%u] = {' % (innertuplename,aggtlistlen)
		tmp_defines1 = '' # for the temp alloc space
		for i in range(aggtlistlen):
			if i != 0:
				tmp_defines2 += ','
			if aggname == 'Avg':
				extraallocname = '%s_extraalloc%i' % (innertuplename, i)
				tmp_defines1 += 'uint8_t %s[%u];\n' % (extraallocname,extraallocsize)
				tmp_defines2 += '{TUPLETYPE_NULL,&%s}' % (extraallocname)
			if aggname == 'Min':
				tmp_defines2 += '{TUPLETYPE_NULL,NULL}'
		tmp_defines2 += '};\n'
		tmp_defines += tmp_defines1
		tmp_defines += tmp_defines2
		
		#write the state var
		if aggname == 'Min':
			tmp_defines += 'AggMinState_t inparam'+str(paramuid)+'_aggstate = {{inparam'+str(paramuid)+'_aggstate_min_gtuple_tlist,0,'+str(hashmap['GT_MAX'])+'}};\n'
		elif aggname == 'Avg':
			tmp_defines += 'AggAvgState_t inparam'+str(paramuid)+'_aggstate = {0,{inparam'+str(paramuid)+'_aggstate_avg_gtuple_tlist,0,'+str(hashmap['GT_MAX'])+'}};\n'
		else:
			templatename = 'Agg%sState_template.nc' % (aggname)
			tmp_instrs += self.factory.instantiate(templatename,hashmap)
			tmp_defines += 'Agg%sState_t inparam%u_aggstate;\n' % (aggname,paramuid)

		return (tmp_defines, tmp_instrs)	


		
		
		
	def writeExprColnum(self,ectuid,attr,colidx,operation):
		defines_add = ''
		instrs_add = ''
		coltype = operation.result.columns[colidx].type
		hashmap = {}
		hashmap['UID'] = ectuid
		hashmap['LOCAL_COL_NUM'] = colidx
		hashmap['LOCAL_COL_TYPE'] = coltype
		hashmap['PID'] = operation.id
		hashmap['pos'] = colidx
		
		if isinstance (attr,il.Index): # todo FIRST it is not il.Index for some reason!
			colnumuid = NescWriter.uid()
			varname = 'col_toget%u_%u' % (operation.id,colidx)
			defines_add += 'uint8_t %s = %u;\n' % (varname, attr.index)
			hashmap['EXPR_TYPE'] = 'EXPR_COL'
			hashmap['TO_GET_POINTER'] = varname
			# dcc todo: to_get_pointer should point to colnumtyped to make it consistent with expr_data
		elif isinstance(attr,list) and len(attr) == 1 and isinstance(attr[0],il.IFunction):	# dcc todo why is function wrapped in a list??
			funcobj = attr[0]
			funcuid = NescWriter.uid()
			varname = 'func_toget%u_%u' % (operation.id,colidx)
			hashmap['EXPR_TYPE'] = 'EXPR_FUNC'
			hashmap['TO_GET_POINTER'] = varname
			(tmp_defines, tmp_instrs) = self.writeFunction(varname,funcobj)
			defines_add += tmp_defines
			instrs_add += tmp_instrs
		else:
			raise RuntimeError, 'sorry, dont know how to handle your attr: %s', attr
		instrs_add += '{' + str(hashmap['EXPR_TYPE'])+',&'+str(hashmap['TO_GET_POINTER'])+',{'+str(hashmap['LOCAL_COL_NUM'])+','+str(hashmap['LOCAL_COL_TYPE'])+'} }'
		return (defines_add,instrs_add)
		
	
		
		
	def writeFunction(self, instancename, funcobj): # funcobj passed in complex
		tmp_defines = ''
		tmp_insts = ''
		functype = None
		for s in self.symboltable.symbols:
			# dcc todo: should be generic symbol table lookup
			if isinstance(s,nesC.Function) and s.name == funcobj.func.typeName:
				functype = s
				break
# 		st()
		if len(functype.parameters) != len(funcobj.param_indexes):
			raise RuntimeError, 'Critical error: function instance parameters not the same length as function type parameters length.  Something bad happened before'
		hashmap = {}
		hashmap['FUNC_NAME'] = instancename
		hashmap['FUNC_TYPE'] = functype.id
		hashmap['FUNC_SIZE'] = len(functype.parameters)
		elistname = '%s_elist' % (instancename)
		hashmap['FUNC_ELIST'] = elistname
		printelist = ''		
		for i,p in enumerate(funcobj.param_indexes):
			# dcc todo: this is really nasty since it does not follow the template model... but our template factory needs to support variable sized arrays... a proper solution is in order...
			if i <> 0:
				printelist += ','
			if isinstance(p,il.Index):
				colname = '%s_param%u' % (elistname,i)
				colnum = p.index
				coltype = functype.parameters[i].type
				printelist += '{EXPR_COL,&'+colname+'}';
				tmp_defines += 'ColnumTyped_t '+colname+' = {'+str(colnum)+','+str(coltype)+'};\n' % ()
			elif isinstance(p,il.Constant):
				# val is ptr
				valname = '%s_param%u' % (elistname,i)
				valtype = functype.parameters[i].type

				# const is referent
				consttype = nesC.getTypeName(valtype)
				constname = '%s_const' % (valname)
				constval = p.constant

				printelist += '{EXPR_VAL,&'+valname+'}';
				tmp_defines += '%s %s = %s;\n' % (consttype,constname,constval)
				tmp_defines += 'ValTyped_t '+valname+' ={&'+constname+','+str(valtype)+'};\n'
			else:
				raise RuntimeError, 'Only single level function nesting supported for now!'
		hashmap['FUNC_ELISTBODY'] = printelist # todo must be a cleaner way to insert array!

		tmp_defines += 'Expr_t '+elistname+'['+str(len(functype.parameters))+'] = {'+printelist+'};\n'
				
		tmp_defines += 'Func_t '+instancename+' = {'+str(hashmap['FUNC_TYPE'])+','+str(hashmap['FUNC_ELIST'])+','+str(hashmap['FUNC_SIZE'])+'};\n'
		return (tmp_defines, tmp_insts)

	def writeRuntime(self):
		hashmap = {}
		hashmap['RuntimeComponentName'] = runtime_component
		#the gtuples
		gtuple_name = 'gtuple_exec_plan'
		gtuples = ''
		gtuples_init = '\t//we do not need it I think..'
		for i,r in enumerate(self.plan):
			gtuples += '\tTuple_t tuple_t_'+str(i)+'['+str(len(r.operations))+'];\n'
			gtuples += '\tGtuple_t '+gtuple_name+'_'+str(i)+' = {'+'&(tuple_t_'+str(i)+'[0]),0,'+str(len(r.operations))+'};\n\n'
		hashmap['GtupleDeclaration'] = gtuples
		hashmap['GtupleInitialization'] = gtuples_init
		maxid = -1
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary:
				continue
			if p.id > maxid:
				maxid = p.id
		hashmap['TTMAX'] = maxid + 1
		
		#now issue the calls open and push
		open_calls_list = ''
		push_calls_list = ''
		copy_temp_gt = ''
		close_calls_list = ''
		for p in self.symboltable.symbols:
			if not isinstance(p,nesC.Predicate):
				continue
			if p.temporary: #or p.primary 
				continue
			case_str = '\n\t\tcase '+str(p.id)+':\n'
			break_str = '\n\t\tbreak;\n'
			open_calls_list += case_str
			push_calls_list += case_str
			copy_temp_gt += case_str
			close_calls_list += case_str
			for i,r in enumerate(self.plan):
				o = r.operations[0]
				open_calls_list += '\t\t//%s\n' % p.name
				if p.id == o.gettables()[0].id :
					open_calls_list += '\t\t\tcall PushOut.open['+str(o.globalid)+']();\n'
					push_calls_list += '\t\t\tcall PushOut.sendnext['+str(o.globalid)+'](&'+gtuple_name+'_'+str(i)+');\n'
					copy_temp_gt += '\t\t'+gtuple_name+'_'+str(i)+'.tlist[0] = temp_gt.tlist[0];\n'
					copy_temp_gt += '\t\t'+gtuple_name+'_'+str(i)+'.size = 1;\n'
					close_calls_list += '\t\t\tcall PushOut.close['+str(o.globalid)+']();\n'
			open_calls_list += break_str
			push_calls_list += break_str
			close_calls_list += break_str
		hashmap['GeneratedOpen'] = open_calls_list
		hashmap['GeneratedPush'] = push_calls_list
		hashmap['CopyTempGT'] = copy_temp_gt
		hashmap['GeneratedClose'] = close_calls_list
		
		output = self.factory.instantiate('RuntimeSyncM_template.nc', hashmap)
		f = self.openOutfile(runtime_component+'.nc')
		print >>f,output

	def writeGenericFunctionCall(self):
		hashmap = {}
		switch_fill = ''
		already_declared = []
		for f in self.symboltable.symbols:
			if isinstance(f,nesC.Function):
				if f.builtin not in already_declared:
					switch_fill += '\n\tcomponents %s;' % (f.builtin)
					already_declared.append(f.builtin)
				switch_fill += '\n\tGenericFunctionCallM.OutCall[%u]->%s;' % (f.id, f.builtin)
				switch_fill += '\n'
		
		hashmap['FunctionWirings'] = switch_fill
		
		output = self.factory.instantiate('GenericFunctionCallC_template.nc', hashmap)
		f = self.openOutfile('GenericFunctionCallC.nc')
		print >>f,output

#NOTE: we assume that all operation classes read from the first table

def generate(symboltable, plan, facts, templatedirs, outdir=None):
	factory = tl.Factory(templatedirs)
	writer = NescWriter(symboltable, plan, facts, factory, outdir)
	writer.writeTables()
	writer.writeWiring()
	writer.writeParams()
	writer.writeRuntime()
	writer.writeGenericFunctionCall()



