#!/usr/bin/python

import copy
import os
import sys
import grammar_classes as gc
from pdb import set_trace as st


# generate and import parser
import yapps2
yapps2.generate('grammar.g')
import grammar

# debugging
def dprint(*args):
	global DEBUG
	if DEBUG:
		for arg in args:
			print arg,
		print		

# default import path
def defaultimportpath():
	pathstrs = os.environ.get('DSNPATH')
	return pathstrs.split(':')

# parse
def justParse(fname,importpath,alreadyimported):
	global DEBUG
	f = open(fname)
	print 'reading %s' % (os.path.abspath(fname))
	alreadyimported.append(os.path.abspath(fname))
	if os.path.dirname(fname) not in importpath:
		importpath.append(os.path.dirname(fname))
	rules = []
	facts = []
	queries = []
	builtins = []
	materialized = []
	priorities = []
	for line in f:
		p = grammar.parse('clause', line)
		if p == None:
			continue
		else:
#			 print 'O:', p
			if isinstance(p,gc.Rule):
				rules.append(p)
			elif isinstance(p,gc.Fact):
				facts.append(p)
			elif isinstance(p,gc.Query):
				queries.append(p)
			elif isinstance(p,gc.PredicateType):
#				 predicateTypes.append(p)
				pass
			elif isinstance(p,gc.Builtin):
				builtins.append(p)
			elif isinstance(p,gc.Materialize):
				materialized.append(p)
			elif isinstance(p,gc.Priority):
				priorities.append(p)
			elif isinstance(p,gc.Import):
				for impfname in p.imports:
					isalreadyimported = False
					filefound = False
					irules = ifacts = iqueries = ibuiltins = imaterialized = ipriorities = None
					# create temporary importpath with default path from env var at end
					completeimportpath = copy.deepcopy(importpath)
					completeimportpath.extend(defaultimportpath())
					for imppname in completeimportpath:
						impfullname = os.path.join(imppname,impfname)
						if impfullname in alreadyimported:
							isalreadyimported = True
							break
						try:
							(irules,ifacts,iqueries,ibuiltins,imaterialized,ipriorities) = justParse(impfullname,importpath,alreadyimported)
						except IOError:
							continue
# 						print 'importing %s' % (impfullname)
						filefound = True
						break # import the very first match
					if isalreadyimported:
						print 'skipping %s' % (impfullname)
						continue
					if filefound:
						rules.extend(irules)
						facts.extend(ifacts)
						queries.extend(iqueries)
						builtins.extend(ibuiltins)
						materialized.extend(imaterialized)
						priorities.extend(ipriorities)
					else:
						raise Exception, 'imported file "%s" not found' % (impfname)
			else:
				raise RuntimeError, 'Unsupported type %s' % p
	return (rules, facts, queries, builtins, materialized, priorities)

# sematic checks
def checkCompareVarsExist(rules):	
	for r in rules:
		lookup = []
		exists = []
		for t in r.tail:
			if isinstance(t,gc.CompareVar) or isinstance(t,gc.CompareCon):
				lookup = lookup + t.getVariables()
			else:
				exists = exists + t.getVariables()
		for v in lookup:
			if v not in exists:
				raise RuntimeError, 'Variable %s not in rule %s except for in comparison' % (v, r)
	
# type inference
def typeInference(rules):
	change = True
	i = 0
	while change:
		i += 1
#		 print
#		 print '-- %i -------------------------------' % i
		change = False
		for r in rules:
#			 print
#			 print 'JOIN'
			if r.joinTypeInfer():
				change = True

def checkAllTypesAssigned(rules, facts, queries):
	def isAllTypesAssigned(rules, facts, queries):
		for r in rules:
			if not(r.isAssigned()): return False
		for f in facts:
			if not(f.isAssigned()): return False
		for q in queries:
			if not(q.isAssigned()): return False
		return True
	def isAllTypesAssigned_old():
		for p in gc.PredicateType.typess:
			if not gc.PredicateType.typess[p].isAssigned(): return False
		for f in gc.FunctionType.typess:
			if not gc.FunctionType.typess[f].isAssigned(): return False
		return True
	if not isAllTypesAssigned(rules, facts, queries):
		raise RuntimeError, 'Not all types assigned'
	pts = gc.PredicateType.typess.values()
	fts = gc.FunctionType.typess.values()
	return (pts,fts)	

# property assignment
def assignProperties(builtins, materializes, priorities):
	for b in builtins:
		t = gc.PredicateType.lookupByName(b.name)
		if t == None:
			t = gc.FunctionType.lookupByName(b.name)
		if t == None:
			continue
# 			raise RuntimeError, '%s lacks matching pred or func type' % (b)
		t.builtin = b.nescC
	for m in materializes:
		t = gc.PredicateType.lookupByName(m.name)
		if t == None:
			raise RuntimeError, '%s lacks matching pred type' % (m)
		if t.builtin != None:
			raise RuntimeError, 'PredicateType cannot be builtin and materialized'
# 		st()
		t.materialize = m
	for p in priorities:
		t = gc.PredicateType.lookupByName(p.name)
		if t == None:
			raise RuntimeError, '%s lacks matching pred type' % (p)
		t.priority = p.priority

# args
def optParse():
	global DEBUG
	from optparse import OptionParser	
	usage = "%prog [options] "
	basename = os.path.basename(sys.argv[0])
	description = """
	decsn parser
	"""
	parser = OptionParser(usage=usage, description=description)
	parser.add_option("-d", "--debug", action="store_true",
					  dest="debug", default=False,
					  help="Print debug messages.")
	parser.disable_interspersed_args()
	options, args =  parser.parse_args()
	DEBUG = options.debug
	if len(args) != 1:
		parser.print_help()
		sys.exit(2)
	return (options, args)

# Parse aka "Main"
def parse(fname):
	importpath = [] # other importpaths?
	alreadyimported = []
	(rules, facts, queries, builtins, materializes, priorities) = justParse(fname, importpath, alreadyimported)

	# assign properties to PredicateType and FunctionType
	assignProperties(builtins, materializes, priorities)

	# iterative type inference
	typeInference(rules)
	(predicateTypes,functionTypes) = checkAllTypesAssigned(rules,facts,queries)

	# semantic check of comparison operators
	checkCompareVarsExist(rules)

	# make the program
	program = gc.Program(rules, facts, queries, predicateTypes, functionTypes, builtins, materializes)
	return program

# Main
if __name__=='__main__':
	(options, args) =  optParse()
	program = parse(args[0])
	
#	 # debugging
 	program.dumpall()

