#!/usr/bin/python

import os
import re
import string

LEFT_DELIM = '$$'
RIGHT_DELIM = '$$'
LEFT_DELIM_R = '\$\$'
RIGHT_DELIM_R = '\$\$'

class Factory:
	GLOBAL_HASHMAP = {} # accessibility needs to be global (module-wide) such that delimrepl() which must only take one parameter can access it

	def __init__(self, path):
		if isinstance(path,list) and sum([isinstance(p,str) for p in path]) == len(path):
			self.path = path
		elif isinstance(path,str):
			self.path = [path]
		else:
			raise RuntimeError, 'path specification is invalid!'

	@staticmethod
	def delimrepl(match):
		key = match.group()[len(LEFT_DELIM):(len(match.group())-len(RIGHT_DELIM))]
		if key in Factory.GLOBAL_HASHMAP:
			return str(Factory.GLOBAL_HASHMAP[key])
		else:
			return match.group()

	def opentemplate(self,fname):
		for p in self.path:
			if os.path.isfile(p + '/' + fname):
				return open(p + '/' + fname)
		raise RuntimeError, '%s not found in %s' % (fname, self.path)
			
	def instantiate(self,fnames,hashmap):
		rets = []
		# do "map()" with additional arg hashmap
		if not isinstance(fnames,list):
			return self._instantiate(fnames,hashmap)
		for fname in fnames:
			rets.append(self._instantiate(fname,hashmap))
		return rets
	def _instantiate(self,fname,hashmap):
		Factory.GLOBAL_HASHMAP = hashmap
		input = self.opentemplate(fname)
		regexp = '(' + LEFT_DELIM_R + '.*?' + RIGHT_DELIM_R + ')'
		output = ''
		for iline in input:
			p = re.compile(regexp)
			oline = p.sub(Factory.delimrepl, iline)
			output += str.rstrip(oline) + '\n'
		return output
      
# Main
if __name__=='__main__':
	# just an example below
	f = Factory(['../generator/templates','helloworld'])
	hashmap = {'TableName':'MyTable',
			   'MAX_CONNECTIONS':'10',
			   'NUM_ROWS':'30',
			   'NUM_COLUMNS':'12',
			   #                'offsetList':'3,2,1,1,3,1',
			   'TABLE_SIZE_IN_BYTES':'30*12',
			   'my_lookup_ID':'3'}
	print f.instantiate('Table_template.nc', hashmap)
    
