#ifndef __RT_TYPES__
#define __RT_TYPES__

/* NOTE: ERROR_EXIT and SERIOUS_DEBUG have moved to SPSimDbg.h in SP */

// TODO NOTE THAT THE FOLLOWING DEFINES HAVE TO BE THE SAME AS THOSE IN THE PYTHON GENERATOR!!!!!!!!
#define JOIN_TYPE 1
#define PROJECT_TYPE 2
#define SELECT_TYPE 3
#define AGGREGATION_TYPE 4

#define TYPE_MASK 0x0F
#define TYPE_INT 0x01
#define TYPE_UINT 0x02
#define TYPE_CHAR 0x04

#define SIZE_MASK 0xF0
#define SIZE_8 0x10
#define SIZE_16 0x20
#define SIZE_32 0x40

#define TYPE_INT8 (TYPE_INT | SIZE_8)
#define TYPE_UINT8 (TYPE_UINT | SIZE_8)
#define TYPE_CHAR8 (TYPE_CHAR | SIZE_8)

#define TYPE_INT16 (TYPE_INT | SIZE_16)
#define TYPE_UINT16 (TYPE_UINT | SIZE_16)
#define TYPE_CHAR16 (TYPE_CHAR | SIZE_16)

#define TYPE_INT32 (TYPE_INT | SIZE_32)
#define TYPE_UINT32 (TYPE_UINT | SIZE_32)
#define TYPE_CHAR32 (TYPE_CHAR | SIZE_32)

#define getTypeID(x) (x | TYPE_MASK)

#define getSizeVal(x) ((x | SIZE_MASK)>> 1)

typedef int8_t Priority_t;
enum {
	MIN_PRIORITY = -128,
};

/* *** Connection IDs *** */
typedef int8_t Connum_t;

enum {
  CONNECTION_NULL = -1,
};

typedef uint16_t Addr_t;

typedef enum {
  GETNEXT_RESULT_FAIL,
  GETNEXT_RESULT_SUCCESS,
  GETNEXT_RESULT_SUCCESS_AND_END,  
} GetnextResult_t;

typedef struct ConnectionContainerTwo_t {
  Connum_t leftComp;
  Connum_t rightComp;
} ConnectionContainerTwo_t;

typedef struct ConnectionContainerThree_t {
  Connum_t leftComp;
  Connum_t rightComp;
  Connum_t topComp;
} ConnectionContainerThree_t;

/* *** Tuples *** */

typedef void TupleData_t;
typedef uint8_t TupleType_t;


typedef struct Tuple_t {
  TupleType_t type;
  TupleData_t *data;
} Tuple_t;


typedef struct Gtuple_t {
  Tuple_t *tlist;
  uint8_t size;
  uint8_t max;
} Gtuple_t;

typedef struct ResultGtuple_t {
  result_t r;
  Gtuple_t* gt;
} ResultGtuple_t;


enum {
  TUPLETYPE_NULL = 254, /* compiler should never issue TupleType of this val  */
};


/* *** General operator parameters *** */

typedef uint8_t Opnum_t;

typedef void InParams_t;

typedef struct Params_t {
  InParams_t *in;
  //Gtuple_t *out;
} Params_t;

typedef struct Restore_t {
  uint8_t insize;
} Restore_t;

/* *** Specific operator parameters *** */

typedef void Val_t;
typedef uint8_t ValType_t;
typedef uint8_t Colnum_t;

typedef struct ValTyped_t {
  Val_t *val;
  ValType_t type;
} ValTyped_t;

typedef struct ColnumTyped_t {
  Colnum_t col;
  ValType_t type;
} ColnumTyped_t;

typedef struct ColnumPairTyped_t {
  Colnum_t col1;
  Colnum_t col2;
  ValType_t type;
} ColnumPairTyped_t;

typedef struct JoinParams_t {
  ColnumPairTyped_t *cplist;	/* points to compiler-generated array */
							/* 	The array specifies a series of pairs of columns, each pair
  							specifying the col1 of tableA and the col2 of tableB on which to
  							join */  
  uint8_t size;
  Gtuple_t* gt;
  Restore_t restore;
	bool negation;
} JoinParams_t;

typedef uint8_t AggType_t;

typedef void AggState_t; /* Look in AggregationTypes.h for AggState_t */
#include "AggregationTypes.h"

typedef struct AggParams_t {
  AggType_t type;
  /* Type is index into function table of  Min, Max, Med, Avg... */
  ColnumTyped_t col; /* todo should be arbitrary Expr_t */
  AggState_t *state;
  Gtuple_t *gt; /* caller-allocated for use with pullin */
  Restore_t restore;
} AggParams_t;

typedef void ExprData_t; /* points to struct indicated by ExprType_t */

typedef enum {
  EXPR_COL = 0,  /* ColnumTyped_t */
  EXPR_VAL = 1,  /* ValTyped_t */
  EXPR_FUNC = 2, /* Func_t */
} ExprType_t;

typedef struct Expr_t {
  ExprType_t type;
  ExprData_t *data;
} Expr_t;

typedef struct ExprColnumTyped_t {
  uint8_t toget_type; //colnum_t
  void* toget;
/*   Expr_t e; */
  ColnumTyped_t c;
} ExprColnumTyped_t;

typedef enum {
  SELTYPE_EQUAL,
  SELTYPE_NOTEQUAL,
  SELTYPE_GREATER,
  SELTYPE_GREATEROREQUAL,
  SELTYPE_LESS,
  SELTYPE_LESSOREQUAL,
} SelType_t;

typedef struct SelParams_t {
/*   CompareResult_t cr; */
  SelType_t type;
  Expr_t lhs;
  Expr_t rhs;
} SelParams_t;

typedef struct ProjParams_t {
  ExprColnumTyped_t *eclist; /* points to compiler-generated array */
  uint8_t size;
} ProjParams_t;

/* *** Expression and function support *** */

typedef uint8_t FuncType_t;

typedef struct Func_t {
  FuncType_t type;
  Expr_t *elist;
  uint8_t size;
} Func_t;

/* ??? typedef void Return_t; /\* allocation of return space *\/ */

/* typedef struct FuncRet_t { /\* like Func_t but with space for return value *\/ */
/*   Func_t func; */
/*   ValTyped_t ret; */
/* } FuncRet_t; */

/* *** VAM *** */

typedef enum {
  COMPARERESULT_NULL,
  COMPARERESULT_GREATER,
  COMPARERESULT_LESS,
  COMPARERESULT_EQUAL,
} CompareResult_t;

#endif /* __RT_TYPES__ */
