#ifndef __TABLE_H__
#define __TABLE_H__

#include "EvictPolicy.h"

// Rownum_t
enum {
	INVALID_ROW = -1, // must be -1 to let first incr go to zero
};
typedef int8_t Rownum_t; // NOT uint8_t to allow for INVALID_ROW assignment

// RowState_t
enum {
	UNUSED_ROW = 0x0000, // default assumption... nothing is actually init'ed
	DATA_USED_ROW = 0x0001,
	FRESH_ROW = 0x0002,
	DELETED_ROW = 0x0004,
	UNDER_CONSTRUCTION_ROW = 0x0008,
	UART_ROW = 0x0010,
};
typedef uint16_t RowState_t;

// ConType_t
enum {
	UNUSED_CONNECTION	= -1,
	REGULAR_CONNECTION = 1,
	FRESH_CONNECTION = 2,
	NON_LOCAL_CONNECTION = 4,
	FRESH_NON_LOCAL_CONNECTION = FRESH_CONNECTION | NON_LOCAL_CONNECTION,
	UART_CONNECTION = 8,
};
typedef int8_t ConType_t;

/* typedef uint8_t Connum_t; */

	// todo (advanced): rearrange columns of tuples so that packing is
	// more efficient (uint8_t + uint8_t on one word) and correct
	// (uint16_t is always byte aligned).  Currently, everything is
	// expanded to uint16_t.

	// todo (questionable): "row_size_in_bytes" could be derived from
	// offset_array[rows_max] if an extra field is added to
	// offset_array. probably would be less space efficient, but easier
	// to read code

#define GARBAGE_COLLECTION_COUNT_THRESHOLD 3

typedef struct TableParams_t {
	// Constant fields
	const TupleType_t tt; // MY_TYPE_ID
	const Connum_t  cons_max; // connections
	const Colnum_t  cols_max; // cols
	const Rownum_t  rows_max; // rows
	const uint8_t   row_size_in_bytes;
	const uint8_t   row_size_in_bytes_unpadded; // minus padding
	const uint8_t   host_col;
	const EvictPolicy_t evict_policy;
	const uint32_t  timeout_value;
	const bool      replacepk; // given a new entry and an existing entry with same primary key, is the existing entry replaced?
	const Priority_t priority; // default priority is 0
	const uint8_t * offset_array; // for each entry the entry i is the offset of the column i in a tuple
	const uint8_t * primary_keys; // bitmap of size (NUM_COLUMNS/8+1)
#ifdef DEBUG
	const DebugMode d; //DebugMode d = {TRUE, DBG_USR2, DbgWhite, "$$TableName$$"};
#endif

	// Variable fields
	uint8_t     busy_entry; // guard for entering stuff in the table; note that we need to prevent access to the temporary buffer on sendnext
	// table state
	uint8_t *	placeholder_to_insert; // temporary storage:[ROW_SIZE_IN_BYTES_UNPADDED]
	uint8_t *	table; 	//the main data sits in here:[ROW_SIZE_IN_BYTES*NUM_ROWS]
	RowState_t *row_state; // refer to row_state enum:[NUM_ROWS]		
	// iterator state
	ConType_t *	connections; // [MAX_CONNECTIONS]
	Rownum_t *	iterators; // [MAX_CONNECTIONS];
	uint8_t *	garbage_collector_counter; //[$$NUM_ROWS$$]
	bool      kicked; // indicates whether or not fact kicker has fired yet

	// conditional code branches
	const bool do_initial_facts_notify;
	const bool do_timeout_timer;
	const bool do_local_notify;
	//cut cause its the same as previous .. should delete in a while const bool everTriggersRule;
} TableParams_t;

#endif /* __TABLE_H__ */
