
package dsntools;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;
import net.tinyos.message.*;

public class Listen2 {
		//// static
		public static void main(String args[]) throws IOException {
				// parse args
				String sfconfig = null;
				String predconfig = null;
				boolean debug = false; 
				for (int i = 0; i < args.length; i++) {
						String arg = args[i];
						if (arg.equals("-sfconfig")) {
								sfconfig = args[++i];
						}
						if (arg.equals("-predconfig")) {
								predconfig = args[++i];
						}
						if (arg.equals("-v")) {
								debug = true;
						}
						if (arg.equals("-h")) {
								printhelp();
								System.exit(0);
						}
				}

				// load connstrs
				List<String> connstrs = null;
				if (sfconfig != null) {
						connstrs = readsfconfig(sfconfig);
				}

				// call main class
				Listen2 x = new Listen2(debug, connstrs, predconfig);
				x.start();
    }

		private static void printhelp() {
				System.out.println("usage: Listen2.java -sfconfig <SFCONFIGFILE> -predconfig <PREDCONFIGFILE> [-debug]");
				System.out.println();
				System.out.println("\t-sfconfig <SFCONFIGFILE>:");
				System.out.println("\t\tfile describing the serial forwarders");
				System.out.println("\t\tconnecting to the network e.g. output");
				System.out.println("\t\tfile generated by masssf.py 'sfconfig.local'");
				System.out.println();
				System.out.println("\t-predconfig <PREDCONFIGFILE>:");
				System.out.println("\t\tfile describing the predicate types ");
				System.out.println("\t\texpected e.g. output file 'gen/.flat_tuplemsg'");

		}

		private static List<String> readsfconfig(String fname) {
				// read input file
				List<String> connstrs = new LinkedList<String>();
				FileInputStream fin;		
				try {
						fin = new FileInputStream(fname);
						BufferedReader in = new BufferedReader(new InputStreamReader(fin));
						String line;
						while ((line = in.readLine()) != null) {
// 								System.out.println(line);
								connstrs.add(line);
						}
						fin.close();

				} catch (IOException e) {
						System.err.println ("Unable to read from file");
						System.exit(-1);
				}
				return connstrs;
		}

    static void threadprint(PrintStream stream, String message) {
        String threadName = Thread.currentThread().getName();
        stream.format("%tT [%s] %s%n", Calendar.getInstance(), threadName, message);
    }

    static void tprint(String message) {
				threadprint(System.out,message);
    }

    static void tprint(Object message) {
				threadprint(System.out,message.toString());
    }

    static void tprint(Exception e) {
				tprint(e.toString());
    }

    static void tprinterr(String message) {
				threadprint(System.err,message);
		}

		static void tprinterr(Exception e) {
				tprinterr(e.toString());
		}

		//// Listen2
		private BlockingQueue<ConnPkt> queue;
		private List<String> connstrs;
		private String predconfigfile;
		private boolean debug;

		public Listen2(boolean _debug, List<String> _connstrs, String _predconfigfile) {
				debug = _debug;
				connstrs = _connstrs;
				predconfigfile = _predconfigfile;
				queue = new LinkedBlockingQueue<ConnPkt>();
// 				connstrs = new LinkedList<String>();
// // 				connstrs.add(null);
// 				// only connect to serial forwarders because can't break readPacket wait from serial
// 				connstrs.add("sf@localhost:7728");
// 				connstrs.add("sf@localhost:7729");
// // 				connstrs.add("serial@COM29:telos");
		}
		
		public void start() {
				tprint("Starting threads");
				List<Endable> endablelist = new ArrayList<Endable>();
				// display
				DisplayRunnable displayrunnable= new DisplayRunnable(debug, predconfigfile, queue);
				Thread displaythread = new Thread(displayrunnable, "DisplayThread");
				endablelist.add(displayrunnable);
				displaythread.start();
				// listen
				for (Iterator<String> it=connstrs.iterator(); it.hasNext();) {
						String connstr = it.next();
						ListenRunnable listenrunnable = new ListenRunnable(debug, connstr, queue);
						String threadname = "ListenThread("+connstr+")";
						Thread listenthread = new Thread(listenrunnable, threadname);
						endablelist.add(listenrunnable);		
						listenthread.start();
				}
				// start
				try {
						Thread.sleep(500); // should really use barrier
						BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
						String str = "";
						System.out.println("Press ENTER to quit");
						str = in.readLine();
				} catch (Exception e) {
						tprinterr(e);
				}
				// end
				for (Iterator<Endable> it = endablelist.iterator(); it.hasNext();) {
						Endable endable = it.next();
						endable.end();
				}
				tprint("I'm done");
		}

		private class ConnPkt {
				private final ListenRunnable conn;
				private final byte[] pkt;
				
				public ConnPkt(ListenRunnable _conn, byte[] _pkt) {
						conn = _conn;
						pkt = _pkt;						
				}

				public byte[] getPkt() {
						return pkt;
				}

				public ListenRunnable getConn() {
						return conn;
				}
		}

		private class ListenRunnable implements Runnable, Endable {
				private final String connstr;
				private final BlockingQueue<ConnPkt> queue;
				private final PacketSource pktsrc;
				private boolean debug;

				public ListenRunnable(boolean _debug, String _connstr, BlockingQueue<ConnPkt> _queue) {
						debug = _debug;
						queue = _queue;
						connstr = _connstr;
						pktsrc = BuildSource.makePacketSource(_connstr);
						if (!checkPktsrc()) {
								return;
						}						
				}

				private boolean checkPktsrc() {
						if (pktsrc == null) {
								tprinterr("Invalid packet source="+connstr);
								return false;
						}
						return true;
				}
				
				public void end() {
						try {
								pktsrc.close();
						} catch (IOException e) {
								tprinterr(e);
						}
				}

				public void run() {
						if (!checkPktsrc()) {
								return;
						}
						try {
								pktsrc.open(PrintStreamMessenger.err);
								tprint("connected");
								while (true) {
										System.out.flush();
										byte[] packet = pktsrc.readPacket();
										ConnPkt connpkt = new ConnPkt(this, packet);
										queue.put(connpkt);
// 										Dump.printPacket(System.out, packet);
// 										System.out.println();
// 										System.out.flush();
								}
						}
						catch (IOException e) {
								tprinterr(e);
						} 
						catch (InterruptedException ex) {
								tprinterr("Interrupted");
						}
				}
		}

		private class DisplayRunnable implements Runnable, Endable {
				private final String predconfigfile;
				private final Map<Integer,PredicateType> id2pt;
				private final BlockingQueue<ConnPkt> queue;
				private Thread self;
				private boolean debug;

				public DisplayRunnable(boolean _debug, String _predconfigfile, BlockingQueue<ConnPkt> _queue) {
						debug = _debug;
						predconfigfile = _predconfigfile;
						id2pt = readpredconfig(_predconfigfile);
// 						id2pt = null;
						queue = _queue;
						self = null;
				}
				public void end() {
						self.interrupt();
				}
				public void run() {
						self = Thread.currentThread();
						try {
								while (true) {
										ConnPkt connpkt  = queue.take();
										byte[] packet = connpkt.getPkt();
										String connstr = connpkt.getConn().connstr;
										if (debug) {
												Dump.printPacket(System.out, packet);
												System.out.println();
										}
										Predicate pred = process(packet);
										// the line below prints the last to chars of the substring which, if using masssf, should be the tosaddr of the mote connected
										//tprint(connstr);
										//tprint("<"+ connstr.substring((connstr.length()-2)) +"> " + pred);
										if (packet[8] == 0x6C) {
										  tprint("<" + getuint16(packet[11], packet[10]) + "> " + pred);
										}
										System.out.flush();
								}
						} catch (InterruptedException ex) {
								tprint("Interrupted");
						}
				}

				private Map<Integer,PredicateType> readpredconfig(String _predconfigfile) {
						// read input file
						tprint("loading " + _predconfigfile);
						Map<Integer,PredicateType> id2pt = new HashMap<Integer,PredicateType>();
						FileInputStream fin;
						try {
								fin = new FileInputStream(_predconfigfile);
								BufferedReader in = new BufferedReader(new InputStreamReader(fin));
								String line;
								while ((line = in.readLine()) != null) {
										PredicateType pt = new PredicateType(line);
										id2pt.put(pt.getid(), pt);
										if (debug) {
												tprint("loaded: " + pt.toString());
										}
								}
								fin.close();
						} catch (Exception e) {
								tprinterr(e);
						}
						return id2pt;
				}

				private Predicate process(byte[] packet) {
						// byte 8 is type
						// byte 9 is group?
						// old way + 
						// byte 10-11 is tosaddr // new way doesn't have tosaddr
						// byte 12 is predicate type
						// byte 13+ is tuple
						// old way -

						if (packet[8] == 0x6C) {
								// todo: not sure these are sign-safe

// 								// old way +
// 								int tosaddr = getuint16(packet[11], packet[10]);
// 								short id = getuint8(packet[12]);
// 								Integer i = new Integer(id);
// 								PredicateType pt = id2pt.get(i);
// 								Predicate p = pt.parse(packet,13);
// 								tprint("<"+tosaddr+"> " + p);
// 								// old way -

								// todo: not sure these are sign-safe

								// new way +
// 								int tosaddr = getuint16(packet[11], packet[10]);
								int start_pos = 12;//10;
								short id = getuint8(packet[start_pos]);
								Integer i = new Integer(id);
								PredicateType pt = id2pt.get(i);
								Predicate p = pt.parse(packet,start_pos+1);
// 								tprint("<"+"> " + p);
								return p;
								// new way -

								

// 								tprint(pt);
						}
						return null;
				}
		}

		short getuint8(byte b) {
				int firstbyte = (0x000000FF & ((int)b));
				return (short)firstbyte;
		}

		int getuint16(byte b1, byte b2) {
				int firstByte = (0x000000FF & ((int)b1));
				int secondByte = (0x000000FF & ((int)b2));
				return (int) (firstByte << 8 | secondByte);
				
		}
		
		long getuint32(byte b1, byte b2, byte b3, byte b4) {
				int firstByte = (0x000000FF & ((int)b1));
        int secondByte = (0x000000FF & ((int)b2));
        int thirdByte = (0x000000FF & ((int)b3));
        int fourthByte = (0x000000FF & ((int)b4));
				return ((long) (firstByte << 24
												| secondByte << 16
                        | thirdByte << 8
                        | fourthByte))
						& 0xFFFFFFFFL;
		}

		private class Predicate {
				private final List<List<Integer>> tuple;
				private final PredicateType type;
				public Predicate(PredicateType _pt, List<List<Integer>> _tuple) {
						type = _pt;
						tuple = _tuple;
				}

				public String toString() {
						StringBuilder sb = new StringBuilder(type.getname() + " ");
						boolean first = true;
						for (Iterator<List<Integer>> it = tuple.iterator(); it.hasNext();) {
								List<Integer> field = it.next();
								if (first) {
										sb.append("(");
										first = false;
								} else {
										sb.append(", ");
								}

								Long val;
								switch(field.size()) {
								case 1:
										val = new Long(getuint8(field.get(0).byteValue()));
										break;
								case 2:
										val = new Long(getuint16(field.get(1).byteValue(),field.get(0).byteValue()));
										break;
								case 4:
										val = new Long(getuint32(field.get(3).byteValue(),field.get(2).byteValue(),field.get(1).byteValue(),field.get(0).byteValue()));
										break;
								default:
										val = null;
										tprinterr("unhandled type...");
										System.exit(1);
								}
								sb.append(val);
// 								sb.append("|" + Long.toHexString(val.longValue()));

						}
						sb.append(")");
						return sb.toString();
				}
		}

		private class PredicateType { // same naming as grammar_classes.py
				private final String name;
				private final Integer id;
				private final List<Integer> offsets;
				public PredicateType(String configstr) throws Exception {
						String[] strs = configstr.split(" ");
						if (strs.length != 3) {
								throw new Exception("malformed configstr");
						}
						name = strs[0];
						id = Integer.valueOf(strs[1]);
						String [] strs2 = strs[2].split(",");
						offsets = new ArrayList<Integer>();
						for (int i = 0; i < strs2.length; i++) {
								offsets.add(Integer.valueOf(strs2[i]));
						}
				}
				public Integer getid() {
						return id;
				}
				public String getname() {
						return name;
				}
				public Predicate parse(byte[] packet, int base) {
						Iterator<Integer> it = offsets.iterator();
						int offset = it.next().intValue(); // ensure at least one
						boolean last = false;
						List<List<Integer>> tuple = new ArrayList<List<Integer>>();
						List<Integer> field = null;
						for (int i = base; i < packet.length; i ++) {
								if (i >= offset + base) {
										if (it.hasNext()) {
												offset = it.next().intValue();
										}
										else {
												offset = packet.length - base;
										}
// 										System.out.println("seeking "+offset);
										if (field != null) {
												tuple.add(field);
										}
										field = new ArrayList<Integer>();
								}
								field.add(new Integer((int)getuint8(packet[i])));
// 								System.out.format("%02x",packet[i]);
// 								System.out.println();
								
						}
						tuple.add(field);
						return new Predicate(this,tuple);
				}

				public String toString() {
						return name +"-" +id.toString();
				}
		}

// 		private static class CtrlThread implements Runnable {
// 				public void run() {
// 						int port = 4444;
// 						try {
// // 								ServerSocket serverSocket = new ServerSocket(port);
// // 								Socket clientSocket = serverSocket.accept();
// 								Socket clientSocket = new Socket("localhost", port);
// 								PrintWriter out = new PrintWriter(clientSocket.getOutputStream(), true);
// 								String inputLine;
// 								BufferedReader in = new BufferedReader(new InputStreamReader(clientSocket.getInputStream()));
// 								while ((inputLine = in.readLine()) != null) {
// 										threadprint(inputLine);
// 										if (inputLine.equals("end")) {
// 												break;
// 										}
// 								}
// 								out.close();
// 								in.close();
// 								clientSocket.close();
// // 								serverSocket.close();
// 						} catch (Exception e) {
// 								threadprinterr(e.toString());
// 								return;
// 						}
// 				}
// 		}


}
