package dsntools;

import java.io.*;
import java.net.*;
import java.util.*;
//import java.util.concurrent.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;
import net.tinyos.message.*;

public class Listen3 {

  
  
  public static void main(String args[]) throws IOException {
    // parse args
    String sfconfig = null;
    for (int i = 0; i < args.length; i++) {
      String arg = args[i];
      if (arg.equals("-sfconfig")) {
        sfconfig = args[++i];
      }
    }

    // Load connstrs
    List connstrs = null;
    if (sfconfig != null) {
      connstrs = readsfconfig(sfconfig);
    }

    Listen3 x = new Listen3(connstrs);
    x.start();
  }

  private static List readsfconfig(String fname) {
    // read input file
    List connstrs = new LinkedList();
    FileInputStream fin;
    try {
      fin = new FileInputStream(fname);
      System.out.println(fname);
      BufferedReader in = new BufferedReader(new InputStreamReader(fin));
      System.out.println("Buffered reader ok");
      String line;
      while ((line = in.readLine()) != null) {
        System.out.println(line);
	connstrs.add(line);
      }
      fin.close();
    } catch (IOException e) {
        System.err.println("Unable to read from file");
	System.exit(-1);
    }
    return connstrs;
  }

  static void threadprint(PrintStream stream, String message) {
    String threadName = Thread.currentThread().getName();
    //stream.printf("%tT [%s] %s%n", Calendar.getInstance(), threadName, message);
    Calendar myCal = Calendar.getInstance();
    stream.println(myCal.getTime() + " " + threadName + " " + message);
  }

  static void packetprint(PrintStream stream, String message) {
    Calendar myCal = Calendar.getInstance();
    stream.println(myCal.getTime() + " " + message);
    outputFile.println(myCal.getTime() + " " + message);
    //stream.printf("%tT %s%n", Calendar.getInstance(), message);
  }
  
  static void tprint(String message) {
    threadprint(System.out, message);
  }

  static void tprint(Object message) {
    threadprint(System.out, message.toString());
  }

  static void tprint(Exception e) {
    tprint(e.toString());
  }

  static void tprinterr(String message) {
    threadprint(System.err, message);
  }

  static void tprinterr(Exception e) {
    tprinterr(e.toString());
  }

  static void DumpPacket(byte[] packet) {
    String myMessage = "";
    for (int i = 0; i < packet.length; i++) {
      String bs = Integer.toHexString(packet[i] & 0xff).toUpperCase();
      if (packet[i] >= 0 && packet[i] < 16)
        myMessage = myMessage + " 0" + bs;
      else
        myMessage = myMessage + " " + bs;
    }
    packetprint(System.out, myMessage);
  }

  
  // Listen3 Variables
  private List connstrs;
  private static PrintWriter outputFile = null;
  
  public Listen3(List _connstrs) {
    connstrs = _connstrs;
    try {
      outputFile = new PrintWriter(new FileOutputStream("output.txt"));
    } catch(FileNotFoundException e) {
        System.out.println("Error opening the file!");
	System.exit(1);
    }
  }

  public void start() {
    tprint("Starting threads");
    List endablelist = new ArrayList();
    
    // listen
    for (Iterator it = connstrs.iterator(); it.hasNext();) {
      String connstr =(String) it.next();
      ListenRunnable listenrunnable = new ListenRunnable(connstr);
      String threadname = "ListenThread("+connstr+")";
      Thread listenthread = new Thread(listenrunnable,threadname);
      endablelist.add(listenrunnable);
      listenthread.start();
    }

    // start
    try {
      Thread.sleep(500);
      BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
      String str = "";
      System.out.println("Press ENTER to quit");
      str = in.readLine();
    } catch (Exception e) {
        tprinterr(e);
    }

    // end
    for (Iterator it = endablelist.iterator(); it.hasNext();) {
      Endable endable = (Endable)it.next();
      endable.end();
    }
    tprint("I'm done");
    outputFile.close();
    System.exit(0);
  }

  private class ListenRunnable implements Runnable, Endable {
    private final String connstr;
    private final PacketSource pktsrc;

    public ListenRunnable(String _connstr) {
      connstr = _connstr;
      pktsrc = BuildSource.makePacketSource(_connstr);

      if (!checkPktsrc()) {
        return;
      }
    }

    private boolean checkPktsrc() {
      if (pktsrc == null) {
        tprinterr("Invalid packet source="+connstr);
	return false;
      }
      return true;
    }

    public void end() {
      try {
        pktsrc.close();
      } catch (IOException e) {
        tprinterr(e);
      }
    }

    public void run() {
      if (!checkPktsrc()) {
        return;
      }
      try {
        pktsrc.open(PrintStreamMessenger.err);
	tprint("connected");
	while(true) {
	  System.out.flush();
	  byte[] packet = pktsrc.readPacket();
	  //tprint(" ");
	  //Dump.printPacket(System.out, packet);
	  DumpPacket(packet);
	  //System.out.println();
	  System.out.flush();
	}
      } catch (IOException e) {
          tprinterr(e);
      } // catch (InterruptedException ex) {
          //tprinterr("Interrupted");
      //}
    }
  }
}
