package dsntools;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;
import net.tinyos.message.*;

public class Listen5MintRoute {
  //static

  public static void main(String args[]) throws IOException {
    //parse args
    String infile = null;
    String predconfig = null;
    for (int i = 0; i < args.length; i++) {
      String arg = args[i];
      if (arg.equals("-infile")) {
        infile = args[++i];
      }
      if (arg.equals("-predconfig")) {
        predconfig = args[++i];
      }
    }

    Listen5MintRoute x = new Listen5MintRoute(infile,predconfig);
    x.start();
  }

  static void threadprint(PrintStream stream, String message) {
    String threadName = Thread.currentThread().getName();
    stream.format("%tT [%s] %s%n", Calendar.getInstance(), threadName, message);
  }

  static void tprint(String message) {
    threadprint(System.out, message);
  }

  static void tprint(Object message) {
    threadprint(System.out, message.toString());
  }

  static void tprint(Exception e) {
    tprint(e.toString());
  }

  static void tprinterr(String message) {
    threadprint(System.err, message);
  }
  
  private String inFile;
  private String predconfigfile;

  public Listen5MintRoute(String _inFile, String _predconfigfile) {
    inFile = _inFile;
    predconfigfile = _predconfigfile;
  }

  public void start() {
    DisplayRunnable displayrunnable = new DisplayRunnable(inFile, predconfigfile);
    Thread displaythread = new Thread(displayrunnable, "DisplayThread");
    displaythread.start();
  }

  private class DisplayRunnable implements Runnable {
    private final String predconfigfile;
    private final String inFile;
    private final Map<Integer,PredicateType> id2pt;
    private FileInputStream myFile;
    private BufferedReader bufReader;
    private Thread self;

    public DisplayRunnable(String _inFile, String _predconfigfile) {
      predconfigfile = _predconfigfile;
      inFile = _inFile;
      try {
        myFile = new FileInputStream(_inFile);
      } catch (FileNotFoundException e) {
          tprint(e);
      }
      //try {
        bufReader = new BufferedReader(new InputStreamReader(myFile));
      /* } /* catch (IOException e) {
          System.out.println("IO exception = " + e);
	  System.exit(-1);
      }*/
      id2pt = readpredconfig(_predconfigfile);
      self = null;
    }

    public void end() {
      self.interrupt();
    }

    public void run() {
      self = Thread.currentThread();
      String currentLine;
      try {
        currentLine = bufReader.readLine();
	while (currentLine != null) {
	  String[] strs = currentLine.split("  ");
	  try {
	    if (strs.length != 2) {
	      throw new Exception("malformed input");
	    }
	  } catch (Exception e) {
	      tprint(e);
	  }
	  process(strs);
	  System.out.flush();
	  currentLine = bufReader.readLine();
	}
        myFile.close();
      } catch (IOException e) {
          System.out.println("IOException + " + e);
      }
      //bufReader.close();
    }

    private Map<Integer,PredicateType> readpredconfig(String _predconfigfile) {
      //read input file
      tprint("loading " + _predconfigfile);
      Map<Integer,PredicateType> id2pt = new HashMap<Integer,PredicateType>();
      FileInputStream fin;
      try {
        fin = new FileInputStream(_predconfigfile);
	BufferedReader in = new BufferedReader(new InputStreamReader(fin));
	String line;
	while ((line = in.readLine()) != null) {
	  PredicateType pt = new PredicateType(line);
	  id2pt.put(pt.getid(), pt);
	}
	fin.close();
      } catch (Exception e) {
          tprint(e);
      }
      return id2pt;
    }

	public static final int EXPECTED_PAYLOAD_SIZE = 8;
    private void process(String[] strs) {
      String[] myPacket = strs[1].split(" ");
      byte[] packet = new byte[myPacket.length];
      for (int z = 0; z < myPacket.length; z++) {
        packet[z] = Integer.valueOf(myPacket[z],16).byteValue();
      }
      if (packet[8] == 0x6C) {
        int tosaddr = getuint16(packet[11], packet[10]);
        if(myPacket.length > 11+EXPECTED_PAYLOAD_SIZE)
      		System.out.println("WARNING: packet is longer then expected ("+myPacket.length+")");
      	if(myPacket.length < 11+EXPECTED_PAYLOAD_SIZE){
      		System.out.println("ERROR: packet is shorter then expected ("+myPacket.length+") ... returning");
      		return;
      	}
		int crtTOSAddr = getuint16(packet[11], packet[10]);
		int source = getuint16(packet[13], packet[12]);
		int parent = getuint16(packet[15], packet[14]);
		int costSeenByParent = getuint16(packet[17], packet[16]);
		int linkQualityToParent = getuint16(packet[19], packet[18]);
		
		System.out.println(strs[0] + " " + "<" + tosaddr + "> " + 
			" src: "+source+" parentAddr: "+parent+" costSeenByParent:"+costSeenByParent+" linkQtoParent:"+linkQualityToParent);
      }
    }
  }

  short getuint8(byte b) {
    int firstbyte = (0x000000FF & ((int)b));
    return (short)firstbyte;
  }

  int getuint16(byte b1, byte b2) {
    int firstByte = (0x000000FF & ((int)b1));
    int secondByte = (0x000000FF & ((int)b2));
    return (int) (firstByte << 8 | secondByte);
  }

  long getuint32(byte b1, byte b2, byte b3, byte b4) {
    int firstByte = (0x000000FF & ((int)b1));
    int secondByte = (0x000000FF & ((int)b2));
    int thirdByte = (0x000000FF & ((int)b3));
    int fourthByte = (0x000000FF & ((int)b4));
    return ((long) (firstByte << 24 | secondByte << 16 | thirdByte << 8 | fourthByte)) & 0xFFFFFFFFL;
  }

  private class Predicate {
    private final List<List<Integer>> tuple;
    private final PredicateType type;
    
    public Predicate(PredicateType _pt, List<List<Integer>> _tuple) {
      type = _pt;
      tuple = _tuple;
    }

    public String toString() {
      StringBuilder sb = new StringBuilder(type.getname() + " ");
      boolean first = true;
      for (Iterator<List<Integer>> it = tuple.iterator(); it.hasNext();) {
        List<Integer> field = it.next();
	if (first) {
	  sb.append("(");
	  first = false;
	}
	else {
	  sb.append(", ");
	}

	Long val;
	switch(field.size()) {
	  case 1:
	    val = new Long(getuint8(field.get(0).byteValue()));
	    break;
	  case 2:
	    val = new Long(getuint16(field.get(1).byteValue(),field.get(0).byteValue()));
	    break;
	  case 4:
	    val = new Long(getuint32(field.get(3).byteValue(), field.get(2).byteValue(), field.get(1).byteValue(), field.get(0).byteValue()));
	    break;
	  default:
	    val = null;
	    tprinterr("unhandled type...");
	    System.exit(-1);
	}
	sb.append(val);
      }
      sb.append(")");
      return sb.toString();
    }
  }

  private class PredicateType {
    private final String name;
    private final Integer id;
    private final List<Integer> offsets;
    
    public PredicateType(String configstr) throws Exception {
      String[] strs = configstr.split(" ");
      if (strs.length != 3) {
        throw new Exception("malformed configstr");
      }
      name = strs[0];
      id = Integer.valueOf(strs[1]);
      String[] strs2 = strs[2].split(",");
      offsets = new ArrayList<Integer>();
      for (int i = 0; i < strs2.length; i++) {
        offsets.add(Integer.valueOf(strs2[i]));
      }
    }

    public Integer getid() {
      return id;
    }

    public String getname() {
      return name;
    }

    public Predicate parse(byte[] packet, int base) {
      Iterator<Integer> it = offsets.iterator();
      int offset = it.next().intValue();
      boolean last = false;
      List<List<Integer>> tuple = new ArrayList<List<Integer>>();
      List<Integer> field = null;
      for (int i = base; i < packet.length; i++) {
        if (i >= offset + base) {
	  if (it.hasNext()) {
	    offset = it.next().intValue();
	  }
	  else {
	    offset = packet.length - base;
	  }

	  if (field != null) {
	    tuple.add(field);
	  }

	  field = new ArrayList<Integer>();
	}
	field.add(new Integer((int)getuint8(packet[i])));
      }
      tuple.add(field);
      return new Predicate(this, tuple);
    }

    public String toString() {
      return name+"-"+id.toString();
    }
  }
}
